# fmt: off

from typing import Callable, Any, Literal

__all__ = [
    'Tup',
    'Rec',
    'Fn',
    'VarScope',
    'SendBytes',
    'EncodeFmt',
    'DecodeFmt',
    'MethodKind',
    'AsyncMode',
    'Options',
    'optstr',
    'strtup',
    'strlist',
    'typetup',
    'record',
    'Name',
    'Ptr',
    'ID',
    'WorldID',
    'LocalRID',
    'GlobalRID',
    'ResourceUID',
    'FrameKey',
    'SystemRID',
    'FrameID',
    'MessageID',
    'FutureID'
]

################################################################################

type optstr       = str | None
type strtup       = tuple[str, ...]
type strlist      = list[str]
type typetup      = tuple[type, ...]
type record       = dict[str, Any]

type Tup[V]       = tuple[V, ...]
type Rec[V]       = dict[str, V]

type Fn[*X, Y]    = Callable[[*X], Y]
type SendBytes    = Callable[[bytes], None]

type VarScope     = Literal['locals', 'globals']
type TermFmt      = Literal['full', 'type', 'json', 'schema', 'raw']
type EncodeFmt    = TermFmt
type DecodeFmt    = TermFmt

type MethodKind   = Literal['instance', 'class', 'static', None]

type AsyncMode    = Literal['coro', 'future', 'sync', None]  # None means 'use server default'

type Options      = dict[str, bool | int | str | float | None]

################################################################################

type Name         = str
type Ptr          = int    # the result of id(...)
type ID           = int    # natural number starting at 0
type UUID         = int    # effectively random integer
type WorldID      = UUID   # this is a UUID based on hash of MAC address
type FrameID      = UUID
type LocalRID     = Ptr
type SystemRID    = ID
type MessageID    = ID
type FutureID     = str | int

################################################################################

# global resource id: the originating world, its frame, and ID there
type GlobalRID    = tuple[WorldID, FrameID, LocalRID]

type ResourceUID  = GlobalRID | SystemRID

# the virtualizing world, its frame, for lookup in servicing requests
type FrameKey     = tuple[WorldID, FrameID]



# fmt: off

from .__ import *

__all__ = [
    'Vars',
]

################################################################################

class Vars:
    __slots__ = __match_args__ = 'vars',

    def __init__(self, vars_: dict):
        self.vars = vars_

    def encode(self, enc: 'EncoderP', fmt: 'EncodeFmt'):
        from ..msg.vars import VarsMsg
        return VarsMsg.encode(enc, self.vars, fmt)



# fmt: off

import asyncio

from agentica_internal.warpc.worlds.debug_world import *


async def verify_virtual_path():


    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair as B:

        path = pair.tmp_file
        path.write_text("data")

        path_b = await B(path)
        assert is_virtual(path_b)
        assert path_b.exists()
        assert path_b.read_bytes() == b"data"
        assert path_b.read_bytes() == b"data"



async def verify_text_write_via_virtual_file():

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair as B:

        path = pair.tmp_file
        with open(path.as_posix(), 'w') as file:
            assert file.write('aaa') == 3
            file_b = await B(file)
            assert is_virtual(file_b)
            assert file.write('xxx') == 3

        assert path.exists()
        assert path.read_text() == 'aaaxxx'


async def verify_text_read_via_virtual_file():

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair as B:

        path = pair.tmp_file
        path.write_text('yyy')
        with open(path.as_posix(), 'r') as file:
            file_b = await B(file)
            assert is_virtual(file_b)
            assert file.readline() == 'yyy'
            assert file.seek(0) == 0
            assert file.readline() == 'yyy'


if __name__ == '__main__':
    asyncio.run(verify_virtual_path())
    asyncio.run(verify_text_write_via_virtual_file())
    asyncio.run(verify_text_read_via_virtual_file())



from .rgb import Rgb

__all__ = [
    'Agentica',
    'Gray',
    'Pure',
    'Light',
    'Medium',
    'Dark',
    'Distinct',
    'Heat',
    'Html',
    'ColorLists',
    'coerce_rgb',
    'named_rgb',
]

####################################################################################################

_hex = Rgb.hex_list


class Agentica:
    light = periwinkle = Rgb.hex('7678ED')
    dark = periwinkle_dark = Rgb.hex('6163EA')


class Gray:
    # fmt: off
    ALL = [Rgb.i(i) for i in range(0, 256, 256 // 15)]
    K0, K1, K2, K3, K4, K5, K6, K7, K8, K9, KA, KB, KC, KD, KE, KF = ALL
    BLACK  = K0
    DARK   = K3
    MEDIUM = K8
    LIGHT  = KC
    WHITE  = KF
    HALF   = Rgb.i(127)
    # fmt: on


class Pure:
    # fmt: off
    white   = Rgb(0xFF, 0xFF, 0xFF)
    black   = Rgb(0x00, 0x00, 0x00)
    gray    = Rgb(0xAA, 0xAA, 0xAA)
    red     = Rgb(0xFF, 0x00, 0x00)
    green   = Rgb(0x00, 0xFF, 0x00)
    blue    = Rgb(0x00, 0x00, 0xFF)
    magenta = Rgb(0xFF, 0x00, 0xFF)
    teal    = Rgb(0x00, 0xFF, 0xFF)
    yellow  = Rgb(0xFF, 0xFF, 0x00)
    ALL = white, black, red, green, blue, magenta, teal, yellow
    # fmt: on


class Light:
    ALL = _hex('ff775e 6caff4 82dd63 ffbb5f bbaff2 7fdbdc 858585 fb77b0 ccffa9')
    R, B, G, O, L, T, K, P, Y = ALL


class Medium:
    ALL = _hex(r'e1432d 3e81c3 4ea82a dc841a 8b7ebe 47a5a7 929292 c74883 f6e259')
    R, B, G, O, L, T, K, P, Y = ALL


class Dark:
    ALL = _hex('b50700 165e9d 217f00 ae5900 665996 0e7c7e 6b6b6b 9e1f61 bba700')
    R, B, G, O, L, T, K, P, Y = ALL


class Distinct:
    # fmt: off
    ALL = _hex('1f77b4 ff7f0e 2ca02c d62728 9467bd 8c564b e377c2 7f7f7f bcbd22 17becf aec7e8 ffbb78 98df8a ff9896 c5b0d5 c49c94 f7b6d2 c7c7c7 dbdb8d 9edae5')
    C1, C2, C3, C4, C5, C6, C7, C8, C9, C10, C11, C12, C13, C14, C15, C16, C17, C18, C19, C20 = ALL
    # fmt: on


class Heat:
    COLD = _hex('f2f5f8 e0e7fc c6d2f9 a4b8f6 809af3 6181f1 4668ef 2e4eee')
    HOT = _hex('fdfdc5 fdfc93 fbf463 f4dc48 eab540 e18d39 d85f31 d0222a')
    C1, C2, C3, C4, C5, C6, C7, C8 = COLD
    H1, H2, H3, H4, H5, H6, H7, H8 = HOT
    ZERO = H0 = C0 = Rgb.h('ffffff')
    ALL = COLD + [ZERO] + HOT


class Html:
    # fmt: off
    _1 = _hex('00ffff f0ffff f5f5dc 000000 0000ff a52a2a ff7f50 dc143c 00ffff ff00ff ffd700 808080 008000 4b0082 f0e68c')
    _2 = _hex('e6e6fa 00ff00 ff00ff 800000 000080 808000 ffa500 da70d6 ffc0cb dda0dd 800080 ff0000 fa8072 c0c0c0')
    _3 = _hex('fffafa d2b48c 008080 ee82ee ffffff ffff00')
    AQUA, AZURE, BEIGE, BLACK, BLUE, BROWN, CORAL, CRIMSON, CYAN, FUCHSIA, GOLD, GRAY, GREEN, INDIGO, KHAKI = _1
    LAVENDER, LIME, MAGENTA, MAROON, NAVY, OLIVE, ORANGE, ORCHID, PINK, PLUM, PURPLE, RED, SALMON, SILVER  = _2
    SNOW, TAN, TEAL, VIOLET, WHITE, YELLOW = _3
    ALL = _1 + _2 + _3
    # fmt: on


# fmt: off
_short: dict[str, str] = {
    'red': 'R', 'blue': 'B', 'green': 'G', 'orange': 'O',
    'lavender': 'L', 'pink': 'L',
    'teal': 'T', 'cyan': 'T', 'purple': 'P', 'yellow': 'Y'
}
# fmt: on


def coerce_rgb(obj: Rgb | str | int | object) -> Rgb:
    if obj is None:
        return Rgb.GRAY
    if isinstance(obj, Rgb):
        return obj
    if isinstance(obj, str):
        if len(obj) == 7 and obj[0] == '#':
            return Rgb.h(obj[1:])
        return named_rgb(obj)
    if isinstance(obj, int) and 0 <= obj < 20:
        return Distinct.ALL[obj]
    print(f'{obj!r} is not an Rgb, color name, or 0..19')
    return Rgb.GRAY


def named_rgb(name: str) -> Rgb:
    family = Medium
    if name.startswith('<'):
        family = Dark
        name = name[1:]
    elif name.startswith('>'):
        family = Light
        name = name[1:]
    name = _short.get(name, name)
    for cls in [Gray, family, Html]:
        dct = cls.__dict__
        if name in dct:
            return dct[name]
    print(f'color: {name!r} not found')
    return Rgb.GRAY


####################################################################################################


class ColorLists:
    # fmt: off
    GRAY     = Gray.ALL
    PURE     = Pure.ALL
    LIGHT    = Light.ALL
    MEDIUM   = Medium.ALL
    DARK     = Dark.ALL
    HTML     = Html.ALL
    DISTINCT = Distinct.ALL
    COLD     = Heat.COLD
    HOT      = Heat.HOT
    HEAT     = Heat.ALL
    HUE      = _hex('e84747 e55f42 e1723b dc8433 d79428 d1a318 c7af01 b5b704 a1be09 8bc40e 70c913 4ccf17 35d035 39cd57 3aca72 36c88c 2cc5a4 10c2bc 37bbca 52b1d5 64a7e0 729dea 7e92f5 8887ff 9b81f9 ab7af3 ba73ed c86ae8 d561e2 df58d6 e455bb e751a1')
    RAINBOW  = _hex('781c86 631d98 4f1faa 472db8 403cc5 3f4ecb 3f5fd0 4270ce 457fcb 4a8cc2 5098b9 57a1ac 5fa99f 68b092 72b584 7db877 89bb6b 96bd60 a2be57 afbd4f bbbc49 c7b944 d1b340 daad3d e0a23a e59738 e68735 e67631 e3612d e04c29 dd3725 db2122')
    # fmt: on



# fmt: off

from datetime import date, datetime, time, timedelta

__all__ = [
    'RICH_OBJS',
]

date_obj      = date(2023, 12, 25)
time_obj      = time(14, 30, 45)
timedelta_obj = timedelta(days=7, hours=3)
datetime_obj  = datetime(2023, 12, 25, 14, 30, 45)

RICH_OBJS = [
    date_obj,
    time_obj,
    timedelta_obj,
    datetime_obj,
]



from agentica_internal.repl.repl import *


def verify_delta_added():
    repl = BaseRepl()
    repl_eval = repl.run_code('''
    x = 5
    ''')
    assert repl_eval.locals_delta.delta_str() == '+x'


def verify_delta_changed():
    repl = BaseRepl()
    repl.update_locals({'x': 1, 'y': 2, 'z': [], 'i': 0})
    repl_eval = repl.run_code('''
    i += 1       # same as i = i + 1
    x = 1        # new is not old
    y = 20       # new is not old
    z = []       # new == old, but is not old
    ''')
    assert repl_eval.locals_delta.delta_str() == '~i ~y ~z'


def verify_delta_removed():
    repl = BaseRepl()
    repl.update_locals({'x': 1, 'y': 2, 'z': 3})
    repl_eval = repl.run_code('''
    del x         # just removed         = -x
    y = 0; del y  # changed then removed = -y  
    q = 0; del q  # added then removed   = nothing
    del z; z = 1  # removed then added   = ~z
    ''')
    assert repl_eval.locals_delta.delta_str() == '-x -y ~z'


if __name__ == '__main__':
    verify_delta_added()
    verify_delta_changed()
    verify_delta_removed()



# fmt: off

from .dclasses import DCLASS_INSTANCES
from .diamonds import DIAMOND_INSTANCES
from .enums import ENUM_INSTANCES
from .named_tuples import NAMED_TUPLE_INSTANCES
from .recursive import RECURSIVE_INSTANCES
from .typed_dicts import TYPED_DICT_INSTANCES
from .units import UNIT_INSTANCES
from .with_annos import WITH_ANNO_INSTANCES
from .with_methods import WITH_METHOD_INSTANCES
from .with_props import WITH_PROP_INSTANCES

__all__ = [
    'DCLASS_INSTANCES',
    'DIAMOND_INSTANCES',
    'ENUM_INSTANCES',
    'NAMED_TUPLE_INSTANCES',
    'RECURSIVE_INSTANCES',
    'TYPED_DICT_INSTANCES',
    'WITH_ANNO_INSTANCES',
    'WITH_METHOD_INSTANCES',
    'WITH_PROP_INSTANCES',
    'INSTANCES'
]

INSTANCES = [
    *DCLASS_INSTANCES,
    *DIAMOND_INSTANCES,
    *ENUM_INSTANCES,
    *NAMED_TUPLE_INSTANCES,
    *RECURSIVE_INSTANCES,
    *TYPED_DICT_INSTANCES,
    *WITH_ANNO_INSTANCES,
    *WITH_METHOD_INSTANCES,
    *WITH_PROP_INSTANCES,
    *UNIT_INSTANCES
]



# fmt: off

from .__ import *
from .base import Msg

__all__ = [
    'LegacyRPCMsg',
    'LegacyResourceRequestMsg',
    'LegacyResourceReplyMsg',
    'LegacyMFIReplyMsg'
]


################################################################################

if TYPE_CHECKING:
    from .resource_def import *
    from .rpc import *
    from .rpc_request_resource import *
    from .rpc_framed import *
    from .rpc_result import *
    from .rpc_sideband import *


################################################################################

class LegacyRPCMsg(Msg):

    mid:   MessageID
    fid:   FrameID

    def upgrade(self) -> 'RPCMsg':
        raise NotImplementedError()

################################################################################

class LegacyResourceRequestMsg(LegacyRPCMsg, tag='rsrc?'):
    """ABC for RPC requests about resources."""

    mid:   MessageID
    fid:   FrameID
    pid:   FrameID
    info: 'ResourceRequestMsg'

    defs: 'Tup[DefinitionMsg]' = ()

    def upgrade(self) -> 'FramedRequestMsg':
        from .rpc_framed import FramedRequestMsg
        return FramedRequestMsg(mid=self.mid, fid=self.fid, data=self.info, defs=self.defs)

################################################################################

class LegacyResourceReplyMsg(LegacyRPCMsg, tag='rsrc!'):

    mid:   MessageID
    fid:   FrameID
    pid:   FrameID
    info: 'ResultMsg'

    defs: 'Tup[DefinitionMsg]' = ()

    def upgrade(self) -> 'FramedResponseMsg':
        from .rpc_framed import FramedResponseMsg
        return FramedResponseMsg(mid=self.mid, fid=self.fid, data=self.info, defs=self.defs)

################################################################################

class LegacyMFIReplyMsg(LegacyRPCMsg, tag='mfi!'):

    mid:   MessageID
    fid:   FrameID
    info: 'ResultMsg'

    defs: 'Tup[DefinitionMsg]' = ()

    @staticmethod
    def make(info: 'ResultMsg') -> 'LegacyMFIReplyMsg':
        return LegacyMFIReplyMsg(mid=-1, fid=0, info=info)

    def upgrade(self) -> 'FutureResultMsg':
        from .rpc_sideband import FutureResultMsg
        return FutureResultMsg(-1, self.info)

# ################################################################################
#
# class LegacyClassDataMsg(Msg, tag='legacycls:'):
#     name:     str
#     cls:     'ResourceMsg'
#     bases:   'ClassesTupleMsg'
#     methods: 'Rec[ResourceMsg]' = ()
#     sattrs:   strtup = ()
#     keys:     strtup = ()
#     annos:   'AnnotationsMsg' = {}
#     attrs:   'AttributesMsg' = {}
#     params:  'ArgsMsg' = ()
#     qname:    optstr = None  # note: same CPython classes don't actually have a qname or module
#     module:   optstr = None
#     doc:      optstr = None
#
#     def upgrade(self) -> 'ClassDataMsg':
#         ClassDataMsg(
#             name=self.name,
#             cls=self.cls,
#             bases=self.bases,
#             methods={k: ('instance', v) for k, v in self.methods.items()},
#             sattrs=self.sattrs,
#             keys=self.keys,
#             annos=self.annos,
#             attrs=self.attrs,
#             params=self.params,
#             qname=self.qname,
#             module=self.module,
#             doc=self.doc,
#         )



# fmt: off

import asyncio
import typing
from typing import NotRequired, Optional, Required, TypedDict

import typing_extensions
from agentica_internal.testing import *
from agentica_internal.testing.examples.classes.enums import *
from agentica_internal.warpc.worlds.debug_world import *


async def verify_typeddict_cls(cls: type[dict]):
    assert typing_extensions.is_typeddict(cls), f"cls is not a typeddict: {cls}"

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair:

        A, B = pair.pipes
        cls_b = await B(cls)
        cls_a = await A(cls_b)

        assert cls_a is cls
        assert cls_b is not cls_a

        assert typing_extensions.is_typeddict(cls_a), f"cls_a is not a typeddict: {cls_a}"
        assert typing_extensions.is_typeddict(cls_b), f"cls_b is not a typeddict: {cls_b}"
        assert type(cls_a) is typing._TypedDictMeta, f"cls_a is not a _TypedDictMeta: {cls_a}"
        assert type(cls_b) is typing._TypedDictMeta, f"cls_b is not a _TypedDictMeta: {cls_b}"

        assert issubclass(cls_a, dict), f"cls_a is not a subclass of dict: {cls_a}"
        assert issubclass(cls_b, dict), f"cls_b is not a subclass of dict: {cls_b}"

        assert len(cls_a.__annotations__) == len(cls_b.__annotations__)
        assert set(cls_a.__annotations__.items()) == set(cls_b.__annotations__.items())

        assert cls_a.__required_keys__ == cls_b.__required_keys__, f"required_keys: {cls_a.__required_keys__} != {cls_b.__required_keys__}"
        assert cls_a.__optional_keys__ == cls_b.__optional_keys__, f"optional_keys: {cls_a.__optional_keys__} != {cls_b.__optional_keys__}"
        assert cls_a.__total__ == cls_b.__total__, f"total: {cls_a.__total__} != {cls_b.__total__}"


def validate_dict_against_typeddict(obj: dict, td_cls: type[dict]) -> None:
    obj_keys = set(obj.keys())
    required_keys = td_cls.__required_keys__
    optional_keys = td_cls.__optional_keys__
    all_allowed_keys = required_keys | optional_keys

    missing_required = required_keys - obj_keys
    assert not missing_required, f"missing required keys: {missing_required}"

    extra_keys = obj_keys - all_allowed_keys
    assert not extra_keys, f"extra keys not in TypedDict: {extra_keys}"

    assert obj_keys <= all_allowed_keys, f"keys {obj_keys - all_allowed_keys} not allowed"

    new_obj = td_cls(**obj)
    assert isinstance(obj, dict)
    assert isinstance(new_obj, dict)
    assert new_obj == obj


async def verify_typeddict_obj(obj: dict):
    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair:
        A, B = pair.pipes

        obj_b = await B(obj)
        obj_a = await A(obj_b)

        assert obj_a == obj
        assert obj_b == obj
        assert obj_a == obj_b

        assert set(obj_a.keys()) == set(obj.keys())
        assert set(obj_b.keys()) == set(obj.keys())

        for key in obj.keys():
            assert obj_a[key] == obj[key]
            assert obj_b[key] == obj[key]


async def verify_typeddict_obj_with_schema(obj: dict, td_cls: type[dict]):
    validate_dict_against_typeddict(obj, td_cls)

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair:
        A, B = pair.pipes

        cls_b = await B(td_cls)

        obj_b = await B(obj)
        obj_a = await A(obj_b)

        validate_dict_against_typeddict(obj, td_cls)
        validate_dict_against_typeddict(obj_a, td_cls)
        validate_dict_against_typeddict(obj_b, cls_b)

        assert obj_a == obj
        assert obj_b == obj



class TypedDict0(TypedDict):
    pass

class TypedDict1(TypedDict):
    a: int
    b: str

TypedDictFunctional = TypedDict("TypedDictFunctional", {"a": int, "b": str})

class TypedDictOptional(TypedDict, total=False):
    a: int
    b: str
    c: float

class TypedDictMixed(TypedDict):
    required_field: str
    optional_field: NotRequired[int]

class TypedDictMixedReverse(TypedDict, total=False):
    optional_field: str
    required_field: Required[int]

class TypedDictWithOptionalType(TypedDict):
    name: str
    age: Optional[int]

class TypedDictNested(TypedDict):
    name: str
    data: dict[str, int]

class TypedDictBase(TypedDict):
    id: int
    name: str

class TypedDictDerived(TypedDictBase):
    email: str
    age: int

class TypedDictWithList(TypedDict):
    items: list[str]
    counts: list[int]

class TypedDictComplex(TypedDict):
    id: int
    name: str
    tags: list[str]
    metadata: dict[str, str]
    score: Optional[float]

TYPEDDICT_CLASSES = [
    TypedDict0,
    TypedDict1,
    TypedDictFunctional,
    TypedDictOptional,
    TypedDictMixed,
    TypedDictMixedReverse,
    TypedDictWithOptionalType,
    TypedDictNested,
    TypedDictBase,
    TypedDictDerived,
    TypedDictWithList,
    TypedDictComplex,
]

TYPEDDICT_INSTANCES = [
    {},
    {"a": 1, "b": "hello"},
    {"a": 42, "b": "world"},
    {"a": 1},
    {"a": 1, "b": "test", "c": 3.14},
    {"required_field": "req"},
    {"required_field": "req", "optional_field": 42},
    {"required_field": 123},
    {"required_field": 456, "optional_field": "opt"},
    {"name": "Alice", "age": 30},
    {"name": "Bob", "age": None},
    {"name": "Test", "data": {"key": 123}},
    {"id": 1, "name": "Base"},
    {"id": 2, "name": "Derived", "email": "test@example.com", "age": 25},
    {"items": ["a", "b", "c"], "counts": [1, 2, 3]},
    {"id": 999, "name": "Complex", "tags": ["tag1", "tag2"], "metadata": {"key1": "value1"}, "score": 95.5},
]

TYPEDDICT_INSTANCES_WITH_SCHEMAS = [
    (TypedDict0(), TypedDict0),
    (TypedDict1(a=1, b="hello"), TypedDict1),
    (TypedDict1(a=999, b="world"), TypedDict1),
    (TypedDictFunctional(a=42, b="functional"), TypedDictFunctional),
    (TypedDictOptional(), TypedDictOptional),
    (TypedDictOptional(a=1), TypedDictOptional),
    (TypedDictOptional(a=1, b="test"), TypedDictOptional),
    (TypedDictOptional(a=1, b="test", c=3.14), TypedDictOptional),
    (TypedDictMixed(required_field="must_have"), TypedDictMixed),
    (TypedDictMixed(required_field="must_have", optional_field=123), TypedDictMixed),
    (TypedDictMixedReverse(required_field=100), TypedDictMixedReverse),
    (TypedDictMixedReverse(required_field=200, optional_field="opt"), TypedDictMixedReverse),
    (TypedDictWithOptionalType(name="Test", age=25), TypedDictWithOptionalType),
    (TypedDictWithOptionalType(name="Test", age=None), TypedDictWithOptionalType),
    (TypedDictNested(name="nested", data={"x": 1, "y": 2}), TypedDictNested),
    (TypedDictBase(id=100, name="base_inst"), TypedDictBase),
    (TypedDictDerived(id=200, name="derived_inst", email="a@b.com", age=30), TypedDictDerived),
    (TypedDictWithList(items=["x", "y"], counts=[10, 20]), TypedDictWithList),
    (TypedDictComplex(id=777, name="complex_inst", tags=["t1", "t2", "t3"], metadata={"m1": "v1", "m2": "v2"}, score=88.8), TypedDictComplex),
    (TypedDictComplex(id=888, name="complex_no_score", tags=["tag"], metadata={}, score=None), TypedDictComplex),
]

INVALID_TYPEDDICT_INSTANCES = [
    ({"a": 1, "b": "test", "extra": "bad"}, TypedDict1, "extra key"),
    ({"a": 1}, TypedDict1, "missing required key 'b'"),
    ({"b": "test"}, TypedDict1, "missing required key 'a'"),
    ({"required_field": "ok", "extra_key": 123}, TypedDictMixed, "extra key"),
    ({}, TypedDict1, "missing all required keys"),
    ({"id": 1}, TypedDictBase, "missing required key 'name'"),
    ({"name": "only name"}, TypedDictBase, "missing required key 'id'"),
    ({"id": 1, "name": "test", "age": 30}, TypedDictBase, "extra key 'age'"),
    ({"optional_field": "opt"}, TypedDictMixedReverse, "missing required field"),
]


def test_validation_catches_errors():
    for invalid_dict, td_cls, description in INVALID_TYPEDDICT_INSTANCES:
        try:
            validate_dict_against_typeddict(invalid_dict, td_cls)
            raise AssertionError(f"validation should have failed for: {description}, dict: {invalid_dict}")
        except AssertionError as e:
            if "validation should have failed" in str(e):
                raise
            pass


def verify_typeddict_cls_sync(cls: type[dict]):
    asyncio.run(verify_typeddict_cls(cls))

def verify_typeddict_obj_sync(obj: dict):
    asyncio.run(verify_typeddict_obj(obj))

def verify_typeddict_obj_with_schema_sync(pair: tuple[dict, type[dict]]):
    obj, td_cls = pair
    asyncio.run(verify_typeddict_obj_with_schema(obj, td_cls))

def verify_typeddict_classes():
    run_object_tests(verify_typeddict_cls_sync, TYPEDDICT_CLASSES)

def verify_typeddict_objects():
    run_object_tests(verify_typeddict_obj_sync, TYPEDDICT_INSTANCES)

def verify_typeddict_objects_with_schemas():
    run_object_tests(verify_typeddict_obj_with_schema_sync, TYPEDDICT_INSTANCES_WITH_SCHEMAS)

if __name__ == '__main__':
    verify_typeddict_classes()
    verify_typeddict_objects()
    verify_typeddict_objects_with_schemas()
    test_validation_catches_errors()



# fmt: off

import asyncio
from types import FunctionType

from agentica_internal.warpc.worlds.debug_world import *


async def verify_warp_as():

    class WarpAs:

        def __call__(self, i: int) -> int:
            return i * 5 + 1

        def ___warp_as___(self):
            return self.__call__

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair:
        A, B = pair.pipes

        warp_as_a = WarpAs()
        warp_as_b = await B(warp_as_a)

        assert is_virtual(warp_as_b)
        assert type(warp_as_b) is FunctionType

        result_a = warp_as_a(100)
        result_b = warp_as_b(100)
        assert result_a == result_b


async def verify_class_warp_as():

    class ClassWarpAs:

        @classmethod
        def ___class_warp_as___(cls):
            return int

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair:
        A, B = pair.pipes

        warp_as_b = await B(ClassWarpAs)

        assert is_real(warp_as_b)
        assert warp_as_b is int


if __name__ == '__main__':

    asyncio.run(verify_warp_as())
    asyncio.run(verify_class_warp_as())



# fmt: off

__all__ = [
    'LOG_GROUPS',
    'GROUP_NAMES',
    'GROUP_EXPAND'
]

###############################################################################

LOG_GROUPS: dict[str, str] = {
    'short': 'send+recv+repl+virt',
    'codec': 'encr+decr',
    'virtual': 'decr+rsrc+virt',
    'core': 'rsrc+repl+virt+ts',
    'server': 'pywasmrunner+sandbox+sessionmanager',
    'sandbox': 'sandbox+pywasmrunner+agentworld+agentrepl',
    'agent': 'agentworld+agentrepl+virt+decr',
    'sdk': 'sdkworld+agent+agenticfunction+virt+frame+csm',
}

GROUP_NAMES = tuple(LOG_GROUPS.keys())
GROUP_EXPAND = LOG_GROUPS.__getitem__



# fmt: off

from .__ import *
from .rpc import RPCMsg


__all__ = [
    'FramedMsg',
    'FramedRequestMsg',
    'FramedResponseMsg',
]


################################################################################

if TYPE_CHECKING:
    from ..request.base import Request
    from .resource_def import DefinitionMsg
    from .rpc_request import RequestMsg
    from .rpc_result import ResultMsg
    from .rpc_legacy import *

################################################################################

class FramedMsg(RPCMsg):
    """ABC for RPC messages that are associated with frames."""

    mid:   MessageID
    fid:   FrameID
    defs: 'Tup[DefinitionMsg]'

    @property
    def thread_name(self) -> str:
        mid = self.mid
        fid = self.fid
        fmt = self.shape
        if mid == -1:
            return fmt
        fmt = f'{mid}:{fmt}'
        if fid == 0:
            return fmt
        return f'{fid}:{fmt}'


################################################################################

class FramedRequestMsg(RPCMsg, tag='req?'):
    """
    Message for all requests associated with a frame.

    Slots:
    * `mid`: the globally unique `MessageID` of this request
    * `fid`: the locally unique `FrameID` of the frame associated with the request
    * `data`: the `RequestMsg` that contains the content of the request
    * `fmt`: the format to encode the response in, one of `'full', 'json', 'schema'
    * `defs`: any auxiliary resource definitions needed to decode the request
    * `async_mode`: if 'coro' or 'future', creates a task to execute the request
    """

    mid:       MessageID
    fid:       FrameID
    data:     'RequestMsg'
    fmt:      'EncodeFmt' = 'full'
    defs:     'Tup[DefinitionMsg]' = ()

    async_mode: AsyncMode = None

    @property
    def is_async(self) -> bool:
        return self.async_mode in ('coro', 'future') or self.data.is_async

    # --------------------------------------------------------------------------

    @property
    def pid(self) -> FrameID:
        return self.fid  # for now!

    # --------------------------------------------------------------------------

    def __shape__(self) -> str:
        return f_id(self.mid) + ', ' + self.data.shape

    def __debug_info_str__(self) -> str:
        return 'mid=' + f_id(self.mid)

    @property
    def thread_name(self) -> str:
        return f'{self.msg_tag}#{f_id(self.mid)}:{self.data.msg_tag}'

    # --------------------------------------------------------------------------

    def decode_request(self, dec: DecoderP) -> 'Request':
        dec_ctx = dec.dec_context(self.defs)
        with dec_ctx:
            request = self.data.decode(dec)
        if self.async_mode:
            return request.set_async_mode(self.async_mode)
        return request

    @staticmethod
    def encode_request(enc: EncoderP, mid: MessageID, fid: FrameID, request: 'Request') -> 'FramedRequestMsg':
        enc_ctx = enc.enc_context()
        with enc_ctx:
            request_msg = request.encode(enc)
        defs = enc_ctx.enc_context_defs()
        async_mode = getattr(request, 'async_mode', None)
        return FramedRequestMsg(
            mid=mid, fid=fid,
            data=request_msg,
            defs=defs,
            async_mode=async_mode,
        )

    def encode_response(self, enc: EncoderP, result: Result) -> 'FramedResponseMsg':
        from .rpc_result import ResultMsg
        enc_ctx = enc.enc_context()
        with enc_ctx:
            result_msg = ResultMsg.encode(enc, result, self.fmt)
        def_msgs = enc_ctx.enc_context_defs()
        return FramedResponseMsg(mid=self.mid, fid=self.fid, data=result_msg, defs=def_msgs)

    def downgrade(self) -> 'LegacyResourceRequestMsg':
        from .rpc_legacy import LegacyResourceRequestMsg
        from .rpc_request_resource import ResourceRequestMsg
        assert isinstance(self.data, ResourceRequestMsg)
        return LegacyResourceRequestMsg(mid=self.mid, fid=self.fid, pid=self.fid, info=self.data)

################################################################################

class FramedResponseMsg(RPCMsg, tag='reply!'):
    """
    A response to a `FramedRequestMsg`.

    Slots:
    * `mid`: the globally unique `MessageID` of this request
    * `fid`: the locally unique `FrameID` of the frame associated with the request
    * `data`: the `ResultMsg` that contains the content of the reply
    * `defs`: any auxiliary resource definitions needed to decode the reply
    """

    fid:   FrameID
    mid:   MessageID
    data: 'ResultMsg'
    defs: 'Tup[DefinitionMsg]' = ()

    @property
    def pid(self) -> FrameID:
        return self.fid  # for now!

    def __shape__(self) -> str:
        return f_id(self.mid) + ', ' + self.data.shape

    def __debug_info_str__(self) -> str:
        return 'mid=' + f_id(self.mid)

    def decode_response(self, dec: DecoderP) -> 'Result':
        dec_ctx = dec.dec_context(self.defs)
        with dec_ctx:
            result = self.data.decode(dec)
        return result

    def downgrade(self) -> 'LegacyResourceReplyMsg':
        from .rpc_legacy import LegacyResourceReplyMsg
        from .rpc_result import ResultMsg
        assert isinstance(self.data, ResultMsg)
        return LegacyResourceReplyMsg(mid=self.mid, fid=self.fid, pid=self.fid, info=self.data)



# fmt: off

from .repl_traceback import sanitize_exception_for_repl

__all__ = [
    'ReplError',
]

################################################################################

class ReplError:
    """
    Represents information about an exception that occurred during `Repl.run_code`.

    The traceback string is set if the exception has a `__traceback__`, it will
    automatically be censored to contain only the part of the traceback that is
    relevant to the agent, avoiding the `run_code` Frame as well as any Frames
    that are warp-related because of virtual resource interaction.
    """

    name:         str
    exception:    BaseException
    traceback:    str | None
    raised:       bool

    def __init__(self, exception: BaseException):
        self.traceback = None
        self.raised = False
        self.set_exception(exception)

    def set_exception(self, exception: BaseException):
        sanitize_exception_for_repl(exception)
        self.exception = exception
        cls = type(self.exception)
        self.name = f"{cls.__module__}.{cls.__qualname__}"

    @property
    def uncaught(self) -> bool:
        return not self.raised

    def __debug_info_str__(self) -> str:
        info = f'name={self.name!r}'
        if self.raised:
            info += f' raised=True'
        return info

    def __short_str__(self):
        info = self.__debug_info_str__()
        return f'<repl_error {info}>'

    __str__ = __short_str__



# fmt: off

from .__ import *
from .rpc import RPCMsg


__all__ = [
    'EventMsg',
    'FutureEventMsg',
    'FutureCanceledMsg',
    'FutureCompletedMsg',
]


################################################################################

if TYPE_CHECKING:
    from .rpc_result import ResultMsg
    from .resource_def import DefinitionMsg
    from ..request.request_future import FutureRequest, CompleteFuture, CancelFuture

################################################################################

class EventMsg(RPCMsg):
    """ABC for RPC events, that do not themselves require responses."""


################################################################################

class FutureEventMsg(EventMsg):
    """ABC for RPC event about futures."""

    future_id: FutureID

    @staticmethod
    def encode_event(enc: EncoderP, future_id: FutureID, result: Result) -> 'FutureEventMsg':
        from .rpc_result import ResultMsg
        if result.is_completed:
            enc_ctx = enc.enc_context()
            with enc_ctx:
                result_msg = ResultMsg.encode(enc, result)
            defs = enc_ctx.enc_context_defs()
            return FutureCompletedMsg(future_id, result_msg, defs)
        elif result.is_canceled:
            return FutureCanceledMsg(future_id)
        else:
            raise E.WarpEncodingError("future result still pending")

    def decode(self, dec: DecoderP) -> 'FutureRequest':
        ...

################################################################################

class FutureCanceledMsg(FutureEventMsg, tag='future_canceled'):
    """Deliver the result of a Future with a given ID (string or integer)."""

    future_id: FutureID

    def decode(self, dec) -> 'CancelFuture':
        from ..request.request_future import CancelFuture
        return CancelFuture(
            dec.future_from_id(self.future_id)
        )

################################################################################

class FutureCompletedMsg(FutureEventMsg, tag='future_completed'):
    """Deliver the result of a Future with a given ID (string or integer)."""

    future_id: FutureID
    result:   'ResultMsg'
    defs:     'Tup[DefinitionMsg]' = ()

    def decode(self, dec) -> 'CompleteFuture':
        from ..request.request_future import CompleteFuture
        with dec.dec_context(self.defs):
            return CompleteFuture(
                dec.future_from_id(self.future_id),
                self.result.decode(dec),
            )



# fmt: off

from .__ import *
from .term import TermMsg, TermPassByValMsg
from ..resource.handle import is_virtual_class


__all__ = [
    'ExceptionMsg',
]


################################################################################

if TYPE_CHECKING:
    from .term_resource import ResourceMsg

################################################################################

# TODO: introduce resources/terms for Tracebacks and such, to unlock powerful remote debugging

EXC_NEW_ARGS = '___exc_new_args___'

class ExceptionMsg(TermPassByValMsg, tag='excp'):

    type V = BaseException

    cls:   'ResourceMsg'
    args:  'Tup[TermMsg]'
    name:   str
    loc:    str | None
    stack:  list[str]

    def __shape__(self) -> str:
        return self.name

    def decode(self, dec: DecoderP) -> V:
        from .term_resource import SystemResourceMsg, UserResourceMsg

        args = dec.dec_sequence(self.args)
        cls_msg = self.cls
        cls: type | None = None
        if isinstance(cls_msg, (SystemResourceMsg, UserResourceMsg)):
            vcls = dec.dec_class(cls_msg)
            is_local = not is_virtual_class(vcls)
            is_exc = isinstance(vcls, type) and issubclass(vcls, BaseException)
            if is_local and is_exc:
                cls = vcls

        if cls is None:
            exc = E.RemoteException(*args, original_cls_msg=cls_msg)
        else:
            exc = cls(*args)
        # if debug and self.loc:
        #     # this is a nicety for internal debugging, but we won't necessarily expose this
        #     exc.add_note(f'Remote location:\n{self.loc}\n')
        # if debug and self.stack:
        #     f_stack = '\n┃ '.join(self.stack)
        #     exc.add_note(f'Remote stack:\n┃ {f_stack}')
        return exc

    @classmethod
    def encode_compound(cls, term: V, enc: EncoderP) -> 'ExceptionMsg':
        from ...cpython.frame import exception_location, exception_stack_strings
        exc = term

        # If this exception originated remotely, prefer the originally encoded
        # class message stored on RemoteException.
        if isinstance(exc, E.RemoteException) and getattr(exc, 'original_cls_msg', None) is not None:
            exc_cls = exc.original_cls_msg
        else:
            exc_cls = enc.enc_class(type(exc))

        if (args := getattr(exc, EXC_NEW_ARGS, None)) is not None:
            exc_args = enc.enc_sequence(args)
        else:
            exc_args = enc.enc_sequence(exc.args)

        exc_name = type(exc).__name__
        exc_loc = exception_location(exc)
        exc_stack = exception_stack_strings(exc)
        return ExceptionMsg(exc_cls, exc_args, exc_name, exc_loc, exc_stack)



from sys import builtin_module_names
from types import ModuleType
from typing import Iterable as Iter
from typing import Literal

__all__ = [
    'module_exports',
    'module_dir',
    'module_items',
    'module_getattrs',
    'module_from_file',
    'module_in_stdlib',
]


MISSING = object()


def module_exports(module: ModuleType) -> list[str] | None:
    """
    Checks a module for a valid `__all__` attribute if present, and returns
    these names in a list.
    """
    assert isinstance(module, ModuleType), f"{module} is not a ModuleType"
    exports = getattr(module, '__all__', None)
    if not isinstance(exports, (list, tuple)):
        return None
    if not all(isinstance(name, str) for name in exports):
        return None
    return list(exports)


type ModOrigin = Literal["module", "submodule", "any"]
type ModGranularity = Literal["modules", "contained"]


def module_dir(module: ModuleType, *, origin: ModOrigin = "submodule") -> list[str]:
    """
    Checks a module for a valid `__all__` attribute if present, and returns
    these names in a list.

    Otherwise, returns the names that `module_items` would emit.
    """

    if (exports := module_exports(module)) is not None:
        return exports

    return [k for k, _ in module_items(module, origin=origin)]


def module_items(
    module: ModuleType,
    *,
    prefer_exports: bool = True,
    origin: ModOrigin = 'submodule',
    granularity: ModGranularity = 'modules',
) -> Iter[tuple[str, object]]:
    """
    Iterates through the namespace of a module and yields `(name, value)` pairs.

    Internal `ModuleType` attributes like, `__path__`, `__doc__`, etc. are skipped.

    `prefer_exports = True` uses `mod.__all__` if available

    `origin = 'module'` ignores objects that aren't defined in this module.
    `origin = 'submodule'` ignores objects that aren't defined in this module or a submodule.
    `origin = 'any'` doesn't ignore any objects.

    `granularity = 'modules'` filters out objects that are modules according to `origin`.
    `granularity = 'contained'` filters out any objects that are contained in a skipped module.
    """
    from agentica_internal.cpython.attrs import ATTRS

    skip = ATTRS.MODULE

    if prefer_exports and (exports := module_exports(module)) is not None:
        for key in exports:
            val = getattr(module, key, MISSING)
            if val is not MISSING:
                yield key, val
        return

    orig_test = ORIG_TESTS[origin]
    get_origin = GET_ORIGIN[granularity]

    mod_name = module.__name__
    for key, val in module.__dict__.items():
        if key in skip:
            continue  # skip internal module attributes

        if val is module:
            continue  # skip the module itself if it happens to be there

        if orig_test and not orig_test(mod_name, get_origin(val)):
            continue

        yield key, val


def get_origin_contained(obj: object) -> str:
    if isinstance(obj, ModuleType):
        return obj.__name__
    name = getattr(obj, '__module__', None)
    if isinstance(name, str):
        return name
    return ''


def get_origin_modules(obj: object) -> str:
    if isinstance(obj, ModuleType):
        return obj.__name__
    return ''


GET_ORIGIN = {'contained': get_origin_contained, 'modules': get_origin_modules}


def origin_equal(mod: str, orig: str):
    return not orig or orig == mod


def origin_within(mod: str, orig: str):
    if not orig or orig == mod:
        return True
    return orig.startswith(mod) and orig[len(mod)] == '.'


ORIG_TESTS = {'any': None, 'module': origin_equal, 'submodule': origin_within}


def module_getattrs(module: ModuleType, attrs: Iter[str]) -> tuple:
    """
    Attempts to get a sequence of attributes from a module all at once.
    Only those that are present are returned -- if any are missing a warning is printed.
    """

    from operator import attrgetter

    try:
        return attrgetter(*attrs)(module)  # type: ignore
    except AttributeError:
        print('failed to get all attributes')
        return tuple(getattr(module, attr) for attr in attrs if hasattr(module, attr))


def module_from_file(file_path: str, mod_name: str) -> ModuleType:
    """
    Creates a module from a single file as a module under the given name.
    """
    import importlib.util

    spec = importlib.util.spec_from_file_location(mod_name, file_path)
    assert spec is not None
    assert spec.loader is not None
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def module_in_stdlib(module_name: str) -> bool:
    if module_name == '__main__':
        return False

    if module_name in builtin_module_names:
        return True

    from stdlib_list import in_stdlib

    return in_stdlib(module_name)



# fmt: off

# from typing import NotRequired, ReadOnly, TypedDict
from typing import NotRequired, TypedDict

from ..shared import *

__all__ = [
    'TypedDictOf0',
    'TypedDictOf1',
    'TypedDictOf2',
    'TypedDictOf3',
    'TypedDictOf2Sub',
    'TypedDictOf3Sub',
    'TypedDictOpt',
#     'TypedDictReadOnly',
    'TypedDictNonTotal',
    'TYPED_DICT_CLASSES',
    'TYPED_DICT_INSTANCES',
]


class TypedDictOf0(TypedDict):
    """TypedDictOf0 docstring"""

    pass


class TypedDictOf1(TypedDict):
    """TypedDictOf1 docstring"""

    a: int


class TypedDictOf2(TypedDict):
    """TypedDictOf2 docstring"""

    a: int
    b: bool


class TypedDictOf3(TypedDict):
    """TypedDictOf3 docstring"""

    a: int
    b: bool
    c: str


class TypedDictOf2Sub(TypedDictOf1):
    """TypedDictOf2Sub docstring"""

    b: bool


class TypedDictOf3Sub(TypedDictOf2Sub):
    """TypedDictOf3Sub docstring"""

    c: str


class TypedDictOpt(TypedDict):
    """TypedDictOpt docstring"""

    a: int
    b: str
    c: NotRequired[str]


# class TypedDictReadOnly(TypedDict):
#     """TypedDictReadOnly docstring"""
#
#     a: int
#     b: str
#     c: ReadOnly[str]


class TypedDictNonTotal(TypedDict, total=False):
    """TypedDictNonTotal docstring"""

    a: int
    b: str
    c: str


TYPED_DICT_CLASSES: list[type] = [
    TypedDictOf0,
    TypedDictOf1,
    TypedDictOf2,
    TypedDictOf3,
    TypedDictOf2Sub,
    TypedDictOf3Sub,
    TypedDictOpt,
    # TypedDictReadOnly,
    TypedDictNonTotal,
]

TYPED_DICT_INSTANCES: list[object] = [
    TypedDictOf0(),
    TypedDictOf1(a=_a),
    TypedDictOf2(a=_a, b=_b),
    TypedDictOf3(a=_a, b=_b, c=_c),  # type: ignore
    TypedDictOf2Sub(a=_a, b=_b),  # type: ignore
    TypedDictOf3Sub(a=_a, b=_b, c=_c),  # type: ignore
]



# fmt: off

from typing import get_args

from ...cpython.classes.anno import *
from .__ import *
from .base import *

__all__ = [
    'TypeData',
    'TypeAliasData',
    'TypeVarData',
    'TypeUnionData',
    'GenericAliasData',
    'CallableTypeData',
    'ForwardRefData',
]

'''
Note, despite the name, 'virtual types' are actually real type annotations
when decoded, and involve no virtual resource requests.
'''

################################################################################

class TypeData(ResourceData):
    __slots__ = ()

    FORBIDDEN_FORM = Any

    @classmethod
    def describe_resource(cls, ty: TypeT) -> 'TypeData':
        return choose_data_class(ty).describe_real_type(ty)

    def create_resource(self, handle: ResourceHandle) -> TypeT:
        handle.kind = Kind.Type
        return self.create_real_type()

    ############################################################################

    @classmethod
    def describe_real_type(cls, ty: TypeT) -> 'TypeData':
        raise NotImplementedError()

    def create_real_type(self) -> TypeT:
        raise NotImplementedError()


def choose_data_class(ty: TypeT) -> type[TypeData]:
    cls = type(ty)
    if cls is TAlias:
        return TypeAliasData
    elif cls is TVar or cls is TParamSpec:
        return TypeVarData
    elif cls is TUnion or cls is CUnion:
        return TypeUnionData
    elif cls is TForward:
        return ForwardRefData
    elif getattr(ty, '__origin__', None) is A.Callable:
        return CallableTypeData
    else:
        return GenericAliasData


################################################################################

class TypeAliasData(TypeData):
    __slots__ = 'name', 'module', 'params', 'value'

    name:   str
    module: str
    params: Tup[TermT]
    value:  TermT

    @classmethod
    def describe_real_type(cls, ty: TypeT) -> 'TypeAliasData':
        assert type(ty) is TAlias

        name = ty.__name__
        module = ty.__module__
        if is_forbidden(ty, module):
            raise E.WarpEncodingForbiddenError(f"<typealias '{module}.{name}'>")

        data = TypeAliasData()
        data.name = name
        data.module = module
        data.params = ty.__type_params__
        data.value = ty.__value__

        return data

    def create_real_type(self) -> TypeT:
        return T.TypeAliasType(self.name, self.value)


################################################################################

class TypeVarData(TypeData):
    __slots__ = 'name', 'pspec'

    name:        str
    pspec:       bool

    @classmethod
    def describe_real_type(cls, ty: TypeT) -> 'TypeVarData':
        assert type(ty) in (TVar, TParamSpec)

        data = TypeVarData()
        data.name = ty.__name__
        data.pspec = isinstance(ty, TParamSpec)

        return data

    def create_real_type(self) -> TypeT:
        if self.pspec:
            return TParamSpec(self.name)
        else:
            return TVar(self.name)


################################################################################

class GenericAliasData(TypeData):
    __slots__ = 'origin', 'args'

    origin:      TermT
    args:        Tup[TermT]

    @classmethod
    def describe_real_type(cls, ty: TGeneric) -> 'GenericAliasData':
        assert type(ty) in GENERICS, f"class {type(ty)} not a generic type class"
        data = GenericAliasData()

        origin = getattr(ty, '__origin__', None)

        # this resolves List[int] back to List
        if type(ty) is TGeneric:
            if name := getattr(ty, '_name', None):
                if name[0].isupper() and hasattr(T, name):
                    origin = getattr(T, name)

        data.origin = origin
        data.args = get_args(ty)
        return data

    def create_real_type(self) -> TypeT:
        origin, args = self.origin, self.args
        if origin is None:
            return Any
        try:
            if type(origin) is TForm and len(args) == 1:
                ty = origin[args[0]]  # e.g. ClassVar[int]
            else:
                ty = origin[args]  # e.g. Tuple[int, str]
        except Exception:
            return Any
        return ty


################################################################################

class CallableTypeData(TypeData):
    __slots__ = 'abc', 'args', 'ret'

    abc:      bool
    args:     TermT | Tup[TermT]
    ret:      TermT

    @classmethod
    def describe_real_type(cls, ty: TypeT) -> 'CallableTypeData':
        origin = getattr(ty, '__origin__', None)
        assert origin is A.Callable, f"wrong origin: {origin} =!= {A.Callable}"
        args, ret = get_args(ty)
        data = CallableTypeData()
        data.abc = type(ty).__module__ == 'collections.abc'
        data.args = tuple(args) if type(args) is list else args
        data.ret = ret
        return data

    def create_real_type(self) -> TypeT:
        args = self.args
        args = list(args) if type(args) is tuple else args
        ret = self.ret
        cls = A.Callable if self.abc else T.Callable
        return cls[args, ret]


################################################################################


class TypeUnionData(TypeData):
    __slots__ = 'alts', 'sys'

    MIGHT_ALIAS = True

    alts:  Tup[TermT]
    sys:   bool

    @classmethod
    def describe_real_type(cls, ty: TypeT) -> 'TypeUnionData':
        assert type(ty) in (CUnion, TUnion)
        data = TypeUnionData()
        data.alts = get_args(ty)
        data.sys = type(ty) is CUnion
        return data

    def create_real_type(self) -> TypeT:
        alts = self.alts
        n = len(alts)
        if n == 0:
            return NoReturn
        if n == 1:
            return alts[0]
        if self.sys:
            t1, *tr = alts
            for t in tr:
                t1 |= t
            return t1
        return T.Union[alts]


################################################################################


class ForwardRefData(TypeData):
    __slots__ = 'name'

    name: str

    @classmethod
    def describe_real_type(cls, ty: TypeT) -> 'ForwardRefData':
        # TODO: should we store module here as well?
        data = ForwardRefData()
        data.name = ty.__forward_arg__ if isinstance(ty, TForward) else ty
        return data

    def create_real_type(self) -> TypeT:
        return TForward(self.name)



# fmt: off

__all__ = [
    'BUILTIN_CLASSES'
]

BUILTIN_CLASSES = [
    int,
    float,
    str,
    bool,
    bytes,
    bytearray,
    type,
    object,
    dict,
    tuple,
    set,
    frozenset,
]



# fmt: off

from enum import EnumMeta

from ..cpython.alias import CALLABLES as FUNC_TYPES
from ..cpython.ids import SYS_MOD_TABLE, SYS_CLASS_ID_DICT, SYS_FUNCTION_ID_DICT, SYS_OBJECT_ID_DICT

from .__ import *
from .forbidden import *

__all__ = [
    'SRID_TO_RSRC',
    'LRID_TO_SRID',
    'SRID_TO_NAME',
    'SYS_MODULES',
    'SYS_EXCEPTIONS',
    'to_system_id',
    'FORBIDDEN_IDS',
]


################################################################################

CLS_OFFSET = 0x00000
OBJ_OFFSET = 0x10000
FUN_OFFSET = 0x20000
MOD_OFFSET = 0x30000
FRB_OFFSET = 0x0FFFF

SYS_MODULES: list[ModuleType] = []
SYS_EXCEPTIONS: list[type[BaseException]] = []

FORBIDDEN_IDS: list[int] = []

################################################################################

def _init_tables():
    global SRID_TO_CLS, SRID_TO_OBJ, SRID_TO_FUN, SRID_TO_MOD
    global SRID_TO_NAME, SRID_TO_RSRC, LRID_TO_SRID
    global SYS_MODULES, SYS_EXCEPTIONS
    global FORBIDDEN_IDS
    from importlib import import_module

    add_exc = SYS_EXCEPTIONS.append
    cls_items, add_cls = mklist()
    obj_items, add_obj = mklist()
    fun_items, add_fun = mklist()
    mod_items, add_mod = mklist()
    rids, add_rid = mklist()
    sids, add_sid = mklist()
    vals, add_val = mklist()
    nams, add_nam = mklist()

    def add(f, val: Any, sid: int, nam: str):
        f((sid, val))
        assert id(val) not in rids, f"DUP: {val} {sid} {nam}"
        add_rid(id(val))
        add_sid(sid)
        add_val(val)
        add_nam(nam)

    get_cls_id = SYS_CLASS_ID_DICT.get
    get_obj_id = SYS_OBJECT_ID_DICT.get
    get_fun_id = SYS_FUNCTION_ID_DICT.get

    cls_i, obj_i, fun_i, mod_i = CLS_OFFSET, OBJ_OFFSET, FUN_OFFSET, MOD_OFFSET

    add(add_obj, None, obj_i + 0, 'None')

    for i, prefix, mod_name in SYS_MOD_TABLE:
        # if mod_name == 'asyncio':
        #     mod_name = '_asyncio'  # hack for now
        mod = import_module(mod_name)
        add(add_mod, mod, mod_i + i, mod_name)
        for name, thing in vars(mod).items():
            if isinstance(thing, type):
                j = get_cls_id(f'{prefix}_{name}')
                if j is None:
                    continue
                add(add_cls, thing, cls_i + j, f'{mod_name}.{name}')
                if issubclass(thing, BaseException):
                    add_exc(thing)
                elif isinstance(thing, EnumMeta):
                    for key, obj in thing._member_map_.items():
                        k = get_obj_id(f'{prefix}_{key}')
                        if k is None:
                            continue
                        add(add_obj, obj, obj_i + k, f'{mod_name}.{name}.{key}')
            elif type(thing) in FUNC_TYPES:
                j = get_fun_id(f'{prefix}_{name}')
                if j is None:
                    continue
                add(add_fun, thing, fun_i + j, f'{mod_name}.{name}')
            elif j := get_obj_id(f'{prefix}_{name}'):
                add(add_obj, thing, obj_i + j, f'{mod_name}.{name}')
            # elif 'coroutine' in name:
            #     print(name, thing, mod_name, type(thing))

    add(add_cls, E.ForbiddenError,  cls_i + 0xFFFD, 'ForbiddenError')
    add(add_cls, E.RemoteException, cls_i + 0xFFFE, 'RemoteException')
    add_exc(E.RemoteException)
    add_exc(E.ForbiddenError)

    frb_id = FRB_OFFSET
    frb_cls_id = cls_i + frb_id
    frb_obj_id = obj_i + frb_id
    frb_mod_id = mod_i + frb_id
    frb_fun_id = fun_i + frb_id

    add(add_cls, forbidden_class,    frb_cls_id, 'forbidden_class')
    add(add_obj, forbidden_object,   frb_obj_id, 'forbidden_object')
    add(add_mod, forbidden_module,   frb_mod_id, 'forbidden_module')
    add(add_fun, forbidden_function, frb_fun_id, 'forbidden_function')
    FORBIDDEN_IDS.extend((frb_cls_id, frb_obj_id, frb_mod_id, frb_fun_id))

    cls_items.sort()
    obj_items.sort()
    fun_items.sort()
    mod_items.sort()

    SRID_TO_CLS = dict(cls_items)
    SRID_TO_OBJ = dict(obj_items)
    SRID_TO_FUN = dict(fun_items)
    SRID_TO_MOD = dict(mod_items)

    SRID_TO_NAME = dict(zip(sids, nams))
    SRID_TO_RSRC = dict(zip(sids, vals))
    LRID_TO_SRID = dict(zip(rids, sids))


################################################################################

SRID_TO_CLS: dict[SystemRID, ClassT] = {}
SRID_TO_OBJ: dict[SystemRID, ObjectT] = {}
SRID_TO_FUN: dict[SystemRID, FunctionT] = {}
SRID_TO_MOD: dict[SystemRID, ModuleT] = {}

SRID_TO_NAME: dict[SystemRID, str] = {}
SRID_TO_RSRC: dict[LocalRID, ResourceT] = {}
LRID_TO_SRID: dict[LocalRID, SystemRID] = {}

def to_system_id(obj: Any) -> SystemRID | None:
    return LRID_TO_SRID.get(id(obj))


################################################################################

_init_tables()
del _init_tables

################################################################################

def print_system_ids():
    print()
    for srid, name in SRID_TO_NAME.items():
        print(f'{srid:8} {srid:6x}   {name}')
    print()

# print_system_ids()



from typing import Any

from agentica_internal.repl.repl import *


def verify_fmt_print():
    class MaklovichRepl(BaseRepl):
        # https://www.youtube.com/watch?v=lIpev8JXJHQ

        def fmt_print(self, args, sep, end) -> str:
            return 'Malkovich? ' + super().fmt_print(args, sep, end)

        def fmt_print_arg(self, value: Any, /) -> str:
            return 'Malkovich!'

    repl = MaklovichRepl(logging=False)
    repl_eval = repl.run_code("print(1, 2, 3)")
    assert repl_eval.output == 'Malkovich? Malkovich! Malkovich! Malkovich!\n'


def verify_fmt_display():
    class AngryRepl(BaseRepl):
        def fmt_display_arg(self, value: Any, /) -> str:
            return repr(value).upper()

    repl = AngryRepl()
    ev = repl.run_code("'ok'")
    assert ev.out_value == 'ok'
    assert ev.out_str == "'OK'"
    assert ev.output == "'OK'\n"

    ev = repl.run_code("def foo(): pass")
    assert not ev.has_out_value


def verify_fmt_exception():
    class MyRepl(BaseRepl):
        def fmt_exception_tb(self, exception, /) -> str:
            return 'an exception occurred'

    repl = MyRepl(logging=False)
    ev = repl.run_code('''
    print(1)
    1 / 0
    print(2)
    ''')
    assert ev.output == '1\nan exception occurred\n'


if __name__ == '__main__':
    verify_fmt_print()
    verify_fmt_display()
    verify_fmt_exception()



# fmt: off

import asyncio
from dataclasses import dataclass

from agentica_internal.warpc.worlds.debug_world import *


@dataclass
class EmptyDataClass:
    """Dataclass with no fields."""
    pass


@dataclass
class DataClass:
    x: int
    y: str
    z: float


async def verify_empty_data_class():

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair:
        A, B = pair.pipes

        a_d = EmptyDataClass()
        b_d = await B(a_d)

        assert is_virtual(b_d)
        b_cls = await B(DataClass)
        assert is_virtual(b_cls)

        a_cls = await A(b_cls)
        assert is_real(a_cls)

    return pair


async def verify_data_class():

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair:
        A, B = pair.pipes

        a_d = DataClass(5, 'abc', 3.5)
        b_d = await B(a_d)

        assert is_virtual(b_d)
        assert a_d.x == b_d.x
        assert a_d.x == b_d.x
        assert a_d.z == b_d.z

        b_cls = await B(DataClass)
        assert is_virtual(b_cls)

        a_cls = await A(b_cls)
        assert is_real(a_cls)

        b_obj = b_cls(5, 'abc', 3.5)
        assert is_virtual(b_obj)

        a_obj = await A(b_obj)
        assert is_real(a_obj)

        assert type(a_obj) is DataClass
        assert a_obj.x == 5
        assert a_obj.y == 'abc'
        assert a_obj.z == 3.5

    return pair


if __name__ == '__main__':
    asyncio.run(verify_empty_data_class())
    asyncio.run(verify_data_class())



# fmt: off

from typing import NoReturn, TYPE_CHECKING

if TYPE_CHECKING:
    from .msg.term_resource import ResourceMsg

__all__ = [
    'WarpError',
    'WarpFrameError',
    'WarpResourceError',
    'WarpProtocolError',
    'WarpCodecError',
    'WarpEncodingForbiddenError',
    'WarpAsyncError',
    'WarpRoleConfusionError',
    'WarpEncodingError',
    'WarpDecodingError',
    'WarpLambdaEncodingError',
    'WarpLambdaDecodingError',
    'WarpShutdown',
    'RemoteException',
    'ForbiddenError',
    'raise_shutdown'
]

################################################################################

class WarpError(BaseException): ...
class WarpFrameError(WarpError): ...
class WarpResourceError(WarpError): ...
class WarpProtocolError(WarpError): ...
class WarpCodecError(WarpError): ...
class WarpEncodingError(WarpCodecError): ...
class WarpDecodingError(WarpCodecError): ...
class WarpEncodingForbiddenError(WarpEncodingError): ...
class WarpAsyncError(WarpEncodingError): ...

class WarpRoleConfusionError(WarpProtocolError):
    """WarpRoleConfusion is raised when an operation is applied to the wrong kind of thing."""

class WarpShutdown(WarpError):
    """
    WarpShutdown is raised inside the python implementation of WasmRunner when a guest thread tries
    to do RPC but the host WasmRunner has been shut down. This terminates the stack that the RPC
    occurred in.
    """

class WarpLambdaEncodingError(WarpEncodingError): ...
class WarpLambdaDecodingError(WarpDecodingError): ...

class RemoteException(RuntimeError):
    original_cls_msg: 'ResourceMsg | None'

    def __init__(self, *args, original_cls_msg: 'ResourceMsg | None' = None):
        super().__init__(*args)
        # Carries the originally encoded exception class message so that passing
        # remote exceptions *back* to the remote can be decoded as the original again.
        self.original_cls_msg = original_cls_msg

class ForbiddenError(RuntimeError): ...

# this is so that warp allows its encoding
# it is also registered in `warpc.system` so it has an SID
RemoteException.__module__ = 'builtins'
ForbiddenError.__module__ = 'builtins'

################################################################################

def raise_shutdown(*args, **kwargs) -> NoReturn:
    raise WarpShutdown()



# fmt: off

from ..transcode import TranscodablePythonMessage
from .__ import *

__all__ = [
    'Msg',
]
################################################################################


def to_tag(s: str):
    return s.removesuffix('Msg').replace('Request', 'Req').replace('Control', 'Ctrl')

class Msg(TranscodablePythonMessage, tag=True, tag_field="msg", frozen=(not TYPE_CHECKING), cache_hash=True,
    omit_defaults=True):
    """
    ABC for all warpc protocol messages.

    Direct subclasses are `RPCMsg`, `RequestMsg`, `ResultMsg`, and `TermMsg`.

    Msgs are cheaply and stably hashable, have a 'tag' field so they form a
    discriminated union.

    ## Properties

    `msg_tag` is the short name used on the wire for a message.

    `shape` gives a short string showing the nested structure of a message,
    using the `msg_tag` as the head, e.g. `request[call[sys[int]]]`. Override
    `__shape__` to customize what the fields appear.

    ## Methods

    `repr()` gives a longer string, with `deep=True` to recurse.
    `pprint()`, `pprint_msgpack()`, `pprint_json()`
    `to_msgpack()`, `from_msgpack()`
    `to_json()`, `from_json()`
    """

    ############################################################################

    # this will be dynamically set to a per-class LogFlag, via mechanism in __final__
    LOG: ClassVar[bool] = False
    TAG: ClassVar[str] = 'msg'
    UNION: ClassVar[T.TypeAliasType]

    # this is populated for every class, in __final__.py, when all classes have been defined
    LEAF_CLASSES: ClassVar[tuple[type['Msg'], ...]] = ()

    ############################################################################

    @property
    def msg_tag(self) -> str:
        return self.__class__.__struct_config__.tag

    @property
    def shape(self) -> str:
        f_args = self.__shape__()
        f_name = type(self).TAG
        if f_args:
            return f'{f_name}[{f_args}]'
        return f_name

    def __shape__(self) -> str:
        return ''

    ############################################################################

    def upgrade(self) -> Self:
        return self

    def downgrade(self) -> Self:
        return self

    ############################################################################

    # hides BLANK fields
    def repr(self, deep: bool = True) -> str:
        f_cls = self.__class__.__name__
        strs, add = mklist()
        nest_fn = repr if deep else f_object
        for k, v in self.__rich_repr__():
            s = f_slot(k, v, nest_fn)
            add(f'{k}={s}' if len(k) > 1 else s)
        return f'{f_cls}({commas(strs)})'

    def __repr__(self) -> str:
        return self.repr()

    def msgpack_str(self, multiline: bool = True) -> str:
        data = self.to_msgpack()
        return fmt_msgpack(data, multiline=multiline)

    ############################################################################

    def pprint(self, err: bool = False):
        P.pprint(self, err=err)

    def pprint_msgpack(self, err: bool = False):
        data = self.to_msgpack()
        pprint_msgpack(data, err=err)

    def pprint_json(self, err: bool = False):
        data = self.to_json()
        pprint_json(data, err=err)

    ############################################################################

    # this is for `core.debug` to use

    def short_str(self) -> str:
        return self.shape

    def __short_str__(self) -> str:
        return self.shape

    # if you want to use the `rich` library to pretty print deep messages
    def __rich_repr__(self):
        fields = self.__class__.__struct_fields__
        fields_1 = [f for f in fields if f.endswith('id')]
        fields_2 = [f for f in fields if f not in fields_1]
        for field in fields_1 + fields_2:
            value = getattr(self, field, UNSET)
            if value is UNSET:
                continue
            if field.endswith('id') and type(value) is int:
                value = f_id(value)
            if field == 'rid' and type(value) is tuple:
                value = f_grid(value)
            yield field, value

    def to_json(self) -> bytes:
        from .__json import enc_json
        return enc_json(self)

    def to_msgpack(self) -> bytes:
        try:
            return enc_msgpack(self)
        except EncodeError as exc:
            P.tprint('error encoding message:', exc, err=True)
            P.tprint('message being encoded:', self, err=True)
            return b''

    @classmethod
    def rec_to_msgpack(cls, rec: dict) -> bytes:
        assert is_rec(rec)
        assert all(isinstance(v, cls) for v in rec.values())
        return enc_msgpack(rec)

    @classmethod
    def from_json[M: Msg](cls: type[M], json_data: bytes) -> M: ...

    @classmethod
    def from_msgpack[M: Msg](cls: type[M], msgpack_data: bytes) -> M: ...

    @classmethod
    def check[M: Msg](cls: type[M], msg: Any):
        assert isinstance(msg, cls), (
            f'expected {cls.__name__!r} message, but got {msg.__class__.__name__}'
        )



# fmt: off

__all__ = [
    'MethodClassBase',
    'MethodClass',
    'MethodClassSub',
    'WITH_METHOD_CLASSES',
    'WITH_METHOD_INSTANCES',
]


class MethodClassBase:
    def __init__(self, init_arg: int):
        """MethodClassBase.__init__ docstring"""
        ...

    def inst_method(self, inst_method_arg: str) -> object:
        """MethodClassBase.inst_method docstring"""
        ...

    @classmethod
    def class_method(cls, class_method_arg: bool) -> object:
        """MethodClassBase.class_method docstring"""
        ...


class MethodClass(MethodClassBase):
    """MethodClass docstring"""

    def __init__(self, init_arg: object):
        """MethodClass.__init__ docstring"""
        ...

    def inst_method(self, inst_method_arg: object) -> str:
        """MethodClass.inst_method docstring"""
        ...

    @staticmethod
    def static_method(static_method_arg: bytes) -> object:
        """MethodClass.static_method docstring"""
        ...

    @classmethod
    def class_method(cls, class_method_arg: int) -> int:
        """MethodClass.class_method docstring"""
        ...


class MethodClassSub(MethodClass):
    """MethodClassSub docstring"""

    def __init__(self, init_arg: object, *int_args, **init_kwargs):
        """MethodClassSub.__init__ docstring"""
        ...

    @classmethod
    def class_method(cls, class_method_arg: object) -> bool:
        """MethodClassSub.class_method docstring"""
        ...


WITH_METHOD_CLASSES: list[type] = [
    MethodClassBase,
    MethodClass,
    MethodClassSub,
]

WITH_METHOD_INSTANCES: list[object] = [
    MethodClassBase(0),
    MethodClass(0),
    MethodClassSub(0),
]



# fmt: off

"""
Module to help customize traceback formatting.
Relatively self-contained (only depends on cpython submodule, not ANSI).
"""

import re
import sys
import typing as T
import shutil

from pathlib import Path
from abc import ABCMeta
from collections.abc import Callable, Collection, Iterable
from os import environ
from os.path import abspath
from sys import __stdout__ as STDOUT, __stderr__ as STDERR
from types import (
    BuiltinFunctionType,
    CoroutineType,
    FrameType,
    FunctionType,
    MethodType,
    NoneType,
    TracebackType,
    UnionType,
)
from typing import NoReturn, TYPE_CHECKING, Any

from .color import Rgb
from .recursion import no_recursion_limit
from ..cpython.frame import *

__all__ = [
    'debug_fmt',
    'fmt_frame',
    'fmt_exception',
    'fmt_traceback',
    'fmt_current_stack',
    'frame_function',
    'frame_location',
    'exception_frame',
    'exception_location',
    'enable_rich_tracebacks',
    'avoid_file',
    'avoid_function',
    'avoid_module',
]

####################################################################################################

if TYPE_CHECKING:
    import asyncio

####################################################################################################

def _print(*args: str, sep: str = ' ', end: str = '\n', err: bool = False) -> None:
    file = STDERR if err else STDOUT
    text = sep.join(a if type(a) is str else '<!str>' for a in args)
    if end:
        text += end
    file.write(text)
    file.flush()

####################################################################################################

type FmtFn = Callable[[Any], str]
type FmtDispatch = dict[type, FmtFn]
type Lines = Iterable[str]

TERM_WIDTH: int = 0
TRIM_LINES: bool = False

IN_DEBUGGER: bool = any(mod in sys.modules for mod in ('pydevd', 'debugpy', 'pdb'))
IN_PYCHARM: bool = 'PYCHARM_HOSTED' in environ
IN_PYTEST: bool = 'PYTEST_VERSION' in environ
COLORTERM: bool = environ.get('COLORTERM') == "truecolor"


def _cls(a: Any) -> type:
    return a.__class__  # type: ignore

GenericAlias         = _cls(list[int])
_GenericAlias        = _cls(T.List[int])
_AnyMeta             = _cls(T.Any)
_SpecialForm         = _cls(T.Self)
_SpecialGenericAlias = _cls(T.List)
_TupleType           = _cls(T.Tuple)
_UnionGenericAlias   = _cls(T.Union[int,str])
_CallableType        = _cls(T.Callable)

TY_CLASSES: tuple[type, ...] = (
    GenericAlias,
    _GenericAlias,
    _AnyMeta,
    _SpecialForm,
    _SpecialGenericAlias,
    _TupleType,
    _UnionGenericAlias,
    _CallableType,
)

TY_SHORT = {
    'GenericAlias':         'Generic',
    'UnionType':            'Union',
    '_SpecialForm':         '_SF',
    '_AnyMeta':             '_AM',
    '_GenericAlias':        '_SGeneric',
    '_SpecialGenericAlias': '_Generic',
    '_UnionGenericAlias':   '_Union',
    '_CallableType':        '_CallableT'
}

dict_keys_t      = type(dict().keys())
dict_values_t    = type(dict().values())
dict_items_t     = type(dict().items())

DICT_TYPES: tuple[type, ...] = dict_keys_t, dict_values_t, dict_items_t

list_iter_t      = type(iter(list()))
tuple_iter_t     = type(iter(tuple()))
set_iter_t       = type(iter(set()))
zip_iter_t       = type(iter(zip((), ())))
map_iter_t       = type(iter(map(hash, ())))

REDUCE_TYPES: tuple[type, ...] = list_iter_t, tuple_iter_t, set_iter_t, zip_iter_t, map_iter_t


class Stringifier(dict[type, FmtFn]):
    ansi: bool
    fmt_object: Callable[..., str]
    fmt_type: FmtFn

    def __init__(self, *args, ansi: bool):
        super().__init__(*args)

        self[list]                 = self.fmt_list
        self[tuple]                = self.fmt_tuple
        self[set]                  = self.fmt_set
        self[dict]                 = self.fmt_dict
        self[GenericAlias]         = self.fmt_galias
        self[UnionType]            = self.fmt_union
        self[_UnionGenericAlias]   = self.fmt_union
        self[_GenericAlias]        = self.fmt_galias
        self[_SpecialGenericAlias] = self.fmt_galias
        self[_SpecialForm]         = self.fmt_sform_ty
        self[_AnyMeta]             = self.fmt_ty_arg
        self[ABCMeta]              = self[type]
        for t in REDUCE_TYPES:
            self[t]                = self.fmt_reduce
        for t in DICT_TYPES:
            self[t]                = self.fmt_dict_part

        self.ansi = ansi
        self.fmt_type = self[type]
        self.fmt_object = self[_object]

    def __call__(self, obj: object) -> str:
        obj_cls = type(obj)
        if fn := self.get(obj_cls):
            return fn(obj)
        for cls, fn in self.items():
            if issubclass(cls, obj_cls):
                return fn(obj)
        return self.fmt_object(obj)

    ############################################################################

    def id_col(self, obj: object) -> FmtFn:
        if not self.ansi:
            return str
        ptr = id(obj)
        return lambda s: _rgb_ptr(ptr, s)

    def fmt_id(self, obj: object, s: str):
        return _rgb_ptr(id(obj), s) if self.ansi else s

    ############################################################################

    def fmt_tuple(self, obj: set) -> str:
        return f'({self.fmt_seq(obj)},)' if obj else '()'

    def fmt_list(self, obj: list) -> str:
        return f'[{self.fmt_seq(obj)}]' if obj else '[]'

    def fmt_set(self, obj: set) -> str:
        return f'set(<{self.fmt_seq(obj)}>)' if obj else 'set()'

    def fmt_dict(self, obj: dict) -> str:
        if not obj:
            return '{}'
        size = len(obj)
        keys = obj.keys
        if size <= 6:
            if all(isinstance(k, str) for k in keys()):
                body = ','.join(f'{k}=…' for k in keys())
                return f'{{{body}}}'
            if all(isinstance(k, (int, bool, str, bytes)) for k in keys()):
                body = ','.join(f'{k!r}:…' for k in keys())
                return f'{{{body}}}'
        return f'{{<{size}>}}'

    def fmt_seq(self, seq: Collection) -> str:
        size = len(seq)
        first = self(next(iter(seq)))
        if size == 1:
            return first
        if size == 2:
            return f'{first}, …'
        return f'{first}, …{size - 1}'

    ############################################################################

    def fmt_reduce(self, obj: object) -> str:
        info = ''
        try:
            _, data, *pos = obj.__reduce__()
            if isinstance(data, tuple) and data:
                info += self.fmt_seq(data)
                if len(pos) == 1:
                    info += 'pos=' + self(pos[0])
        except:
            pass
        return self.fmt_object(obj, info)  # type: ignore

    def fmt_dict_part(self, obj: object) -> str:
        info = ''
        try:
            mapping = getattr(obj, 'mapping', None)
            if mapping is not None:
                info = self(mapping)
        except:
            pass
        return self.fmt_object(obj, info)  # type: ignore

    ############################################################################

    def fmt_call(self, func: str, args: list[tuple[str, object]], k_fn: FmtFn) -> str:
        if len(args) > 6:
            f_items = ''.join(f'\n  {k_fn(k)}={self(v)},' for k, v in args)
            return f'{func}({f_items}\n)'
        f_items = COMMA(f'{k_fn(k)}={self(v)}' for k, v in args)
        return f'{func}({f_items})'

    ############################################################################

    def fmt_union(self, obj: UnionType | _UnionGenericAlias) -> str:  # type: ignore
        info = ''
        try:
            if len(args := obj.__args__) > 1:
                info = self.fmt_ty_args(args, ' | ')
        except Exception:
            pass
        col = self.id_col(obj)
        cname = 'Union' if obj.__class__ is UnionType else '_Union'
        if not info:
            return col(f'<{cname}>')
        return col(f'{cname}[') + info + col(']')

    def fmt_galias(self, obj: GenericAlias | _GenericAlias) -> str:  # type: ignore
        info = ''
        try:
            if hasattr(obj, '__origin__'):
                orig = self.fmt_ty_arg(obj.__origin__)
                args = obj.__args__
                args = self.fmt_ty_args(args, ', ')
                info = f'{orig}[{args}]'
        except Exception:
            pass
        col = self.id_col(obj)
        cname = obj.__class__.__name__
        cname = TY_SHORT.get(cname, cname)
        if not info:
            return col(f'<{cname}>')
        return col(f'{cname}(') + info + col(')')

    def fmt_sform_ty(self, obj: _SpecialForm) -> str:  # type: ignore
        col = self.id_col(obj)
        return col('_SF(') + obj._name + col(')')

    def fmt_ty_args(self, args: tuple, sep: str = ' ') -> str:
        ty_arg = self.fmt_ty_arg
        if args.__class__ is not tuple:
            return '???'
        strs = []
        add = strs.append
        for i, arg in enumerate(args):
            if i >= 2:
                break
            f_arg = ty_arg(arg)
            add(f_arg)
            break
        rem = len(args) - len(strs)
        add(f'…{rem}' if rem > 1 else '…')
        return sep.join(strs)

    def fmt_ty_arg(self, arg: object) -> str:
        cls = arg.__class__
        if cls is type or cls is ABCMeta:
            return self(arg)
        elif issubclass(cls, TY_CLASSES):
            cname = cls.__name__
            cname = TY_SHORT.get(cname) or cname
            return self.fmt_id(arg, cname)
        elif cls is _AnyMeta:
            return self.fmt_id(arg, 'Any')
        else:
            _print(str(arg), end='\n\n')
            return '…'


################################################################################


class DebugFormatter:

    root_path:       str
    avoided_files:   set[str]
    avoided_funcs:   set[int | str]
    fmt_plain:       Stringifier
    fmt_ansi:        Stringifier
    seen:            set[int]


    def __init__(self, *, root_path: str, ansi: bool = True):
        self.root_path = abspath(root_path)
        self.avoided_files = set()
        self.avoided_func_names = set()
        self.avoided_funcs = set()
        self.fmt_plain = Stringifier(PLAIN_DISPATCH, ansi=False)
        self.fmt_ansi = Stringifier(ANSI_DISPATCH, ansi=True)
        self.ansi = ansi and (COLORTERM or not IN_DEBUGGER)
        self.seen = set()

    @property
    def fmt(self) -> Stringifier:
        return self.fmt_ansi if self.ansi else self.fmt_plain

    def fmt_any(self, arg: Any) -> str:
        return self.fmt_ansi(arg) if self.ansi else self.fmt_plain(arg)

    def fmt_args(self, *args) -> str:
        try:
            return ' '.join(map(self.fmt_any, args))
        except Exception as exc:
            return f'<fmt_args: internal error {type(exc).__name__}>'

    def fmt_msg(self, msg: str, *args, **kwargs) -> str:
        fmt = self.fmt
        f_args = [fmt(v) for v in args]
        f_kwargs = {k: fmt(v) for k, v in kwargs.items()}
        return msg.format(*f_args, **f_kwargs)

    def avoid_module(self, module_name: str):
        from pathlib import Path

        if module_name not in sys.modules:
            return
        path = ''
        module = sys.modules[module_name]
        if hasattr(module, '__file__') and module.__file__:
            mod_path = module.__file__
        elif hasattr(module, '__path__'):  # package
            mod_path = module.__path__[0]
        else:
            return
        mod_dir = Path(mod_path).parent
        self.avoided_files.add(mod_dir.as_posix())

    def avoid_file(self, filepath: str):
        self.avoided_files.add(abspath(filepath))

    def avoid_func(self, func: Callable | str):
        fn = func
        add = self.avoided_funcs.add
        if isinstance(func, str):
            add(func)
            return
        while True:
            if isinstance(fn, MethodType):
                add(id(fn))
                fn = fn.__func__
            # elif hasattr(fn, '__wrapped__'):
            #     add(id(fn))
            #     fn = fn.__wrapped__
            elif isinstance(fn, FunctionType):
                add(id(fn))
                break
            else:
                raise TypeError(f'Cannot avoid {func!r}')

    def add_formatter(self, fmt_cls: type, plain_fn: FmtFn, ansi_fn: FmtFn | None = None):
        self.fmt_plain[fmt_cls] = plain_fn
        self.fmt_ansi[fmt_cls] = ansi_fn if ansi_fn else plain_fn

    def see(self, obj: object) -> bool:
        ptr = id(obj)
        if ptr in self.seen:
            return True
        self.seen.add(ptr)
        return False

    ############################################################################

    def use_for_uncaught_exceptions(self) -> None:
        sys.excepthook = self.excepthook

    def excepthook(
        self,
        exc_type: type[BaseException],
        exc_value: BaseException,
        exc_traceback: TracebackType | None,
    ):
        with no_recursion_limit:
            try:
                self.print_caught_exception(exc_value)
            except Exception as fmt_ex:
                fmt_ex_cls = fmt_ex.__class__.__name__
                _print(f'ErrorFormatter.excepthook: error formatting traceback for {fmt_ex_cls}')
                _print('resorting to built-in system excepthook')
                sys.__excepthook__(exc_type, exc_value, exc_traceback)

    ############################################################################

    def print_caught_exception(self, ex: BaseException) -> None:
        with no_recursion_limit:
            style = _EXC_SYS if self.ansi else str
            width = TERM_WIDTH or term_width()
            f_exc = self.fmt_exception(ex, width)
            top = style(justified('UNCAUGHT EXCEPTION', '┏━┓', width))
            bot = style(justified('', '┗━┛', width))
            _print('\n', top, '\n', f_exc, '\n', bot, sep='')

    def panic(self, *args, skip: int = 0, width: int = 0) -> NoReturn:
        with no_recursion_limit:
            try:
                frame = sys._getframe(skip + 1)
                width = width or TERM_WIDTH or term_width()
                f_args = (self.fmt_args(*args), '')
                f_stack = self.fmt_stack(frame, width)
                _print(PANIC_BANNER, *f_args, _PLINE, f_stack, _SLINE, sep='\n', err=True)
                STDOUT.flush()
                STDERR.flush()
            except Exception:
                STDERR.write('panic: internal error while dumping stack\n')
                STDOUT.flush()
            try:
                from os import _exit

                _exit(999)
            except:
                exit(999)

    def print_current_stack(self, *args, skip: int = 0, width: int = 0) -> None:
        with no_recursion_limit:
            try:
                frame = sys._getframe(skip + 1)
                width = width or TERM_WIDTH or term_width()
                f_title = _WARN('CURRENT STACK')
                f_args = self.fmt_args(*args)
                f_stack = self.fmt_stack(frame, width)
                _print(f_title, _SLINE, f_args, f_stack, _SLINE, sep='\n')
            except Exception:
                STDERR.write('print_current_stack: internal error\n')
                STDOUT.flush()

    def print_current_frame(self) -> None:
        with no_recursion_limit:
            frame = sys._getframe(1)
            _print(self.fmt_frame(frame, term_width()))

    def print_all_tasks(self) -> None:
        with no_recursion_limit:
            _print(self.fmt_all_tasks())

    def print_all_threads(self) -> None:
        with no_recursion_limit:
            _print(self.fmt_all_threads())

    def print_tasks(self, task: 'asyncio.Task') -> None:
        with no_recursion_limit:
            _print(self.fmt_task(task))

    def print_coro(self, coro: CoroutineType) -> None:
        with no_recursion_limit:
            _print(self.fmt_coro(coro))

    def print_exception(self, ex: BaseException) -> None:
        with no_recursion_limit:
            self._print_lines(self.exc_lines(ex, term_width()))

    def print_traceback(self, tb: TracebackType) -> None:
        with no_recursion_limit:
            self._print_lines(self.tb_lines(tb, term_width()))

    def print_warning(self, msg: str) -> None:
        with no_recursion_limit:
            msg = f'warning: {msg}'
            _print(_WARN(msg) if self.ansi else msg)

    def _print_lines(self, lines: Lines) -> None:
        w = TERM_WIDTH or term_width()
        lst = []
        add_line = lst.append
        hline = '━' * (w - 2)
        add_line('┏' + hline + '┓')
        for line in lines:
            add_line(line)
        add_line('┗' + hline + '┛')
        sys.stdout.flush()
        _print(*lst, sep='\n')
        sys.stdout.flush()

    ############################################################################

    def push_ansi(self, ansi: bool | None) -> None:
        self._ansi = self.ansi
        if ansi is not None:
            self.ansi = ansi

    def pop_ansi(self) -> None:
        self.ansi = self._ansi

    def fmt_exception(self, ex: BaseException, w: int = 0, ansi: bool | None = None) -> str:
        with no_recursion_limit:
            if not isinstance(ex, BaseException):
                return '<fmt_exception: !BaseException>'
            try:
                self.push_ansi(ansi)
                return self._trim_lines(self.exc_lines(ex, w), w)
            except Exception as exc:
                return f'<fmt_exception: internal error: {type(exc).__name__}>'
            finally:
                self.pop_ansi()

    def fmt_traceback(self, tb: TracebackType, w: int = 0, ansi: bool | None = None) -> str:
        with no_recursion_limit:
            if type(tb) is not TracebackType:
                return '<fmt_traceback: !TracebackType>'
            _ansi = self.ansi
            try:
                self.push_ansi(ansi)
                return self._trim_lines(self.tb_lines(tb, w), w)
            except Exception as exc:
                return f'<fmt_traceback: internal error: {type(exc).__name__}>'
            finally:
                self.pop_ansi()

    def fmt_current_stack(self, *args, skip: int = 0, width: int = 0) -> str:
        with no_recursion_limit:
            frame = sys._getframe(skip + 1)
            f_frame = self.fmt_stack(frame, width)
            if not args:
                return f_frame
            f_args = self.fmt_args(*args)
            return join_nl((_PLINE, f_args, _PLINE, f_frame))

    def fmt_stack(self, frame: FrameType, w: int = 0) -> str:
        with no_recursion_limit:
            if not isinstance(frame, FrameType):
                return '<fmt_stack: !FrameType>'
            try:
                return self._trim_lines(self.stack_lines(frame, w), w)
            except Exception as exc:
                return f'<fmt_stack: internal error: {type(exc).__name__}>'

    def fmt_frame(self, frame: FrameType, w: int = 0) -> str:
        with no_recursion_limit:
            if not isinstance(frame, FrameType):
                return '<fmt_frame: !FrameType>'
            try:
                loc, fun = self._fmt_frame(frame)
                sty = _LOC if self.ansi else str
                trim = self._trim
                if w and len(loc) > w / 2:
                    loc = trim(loc, w)
                    fun = trim(fun, w)
                    return f'{_CALL1}{fun}\n{_CALL2}{sty(loc)}'
                return trim(f'{_CALL1}{sty(loc)} {fun}', w)
            except Exception as exc:
                return f'<fmt_frame: internal error: {type(exc).__name__}>'

    def fmt_task(self, task: 'asyncio.Task', w: int = 0) -> str:
        with no_recursion_limit:
            try:
                return self._trim_lines(self.task_lines(task, w), w)
            except Exception as exc:
                return f'<fmt_task: internal error: {type(exc).__name__}>'

    def fmt_coro(self, coro: CoroutineType, w: int = 0, avoid: str = '') -> str:
        with no_recursion_limit:
            try:
                return self._trim_lines(self.coro_lines(coro, w), w)
            except Exception as exc:
                return f'<fmt_coro: internal error: {type(exc).__name__}>'

    def fmt_all_tasks(self, w: int = 0) -> str:
        with no_recursion_limit:
            from asyncio import all_tasks

            lines = []
            extend = lines.extend
            for task in all_tasks():
                # sys.__stderr__.write('\n')
                # task.print_stack(file=sys.__stderr__)
                # sys.__stderr__.write('\n')
                # sys.__stderr__.flush()
                extend(self.task_lines(task, w))
                extend([''])
            return join_nl(lines)

    def fmt_all_threads(self, w: int = 0) -> str:
        with no_recursion_limit:
            import threading

            lines = []
            for thread in threading.enumerate():
                name, ident, alive, daemon = (
                    thread.name,
                    thread.ident,
                    thread.is_alive(),
                    thread.daemon,
                )
                args = [f'{name=!r}']
                if ident:
                    args.append(f'{ident=!r}')
                if not alive:
                    args.append('alive=False')
                if daemon:
                    args.append('daemon=True')
                lines.append('Thread(' + ','.join(args) + ')')
            return join_nl(lines)

    ############################################################################

    def _trim(self, text: str, w: int = 0) -> str:
        if w == 0 or not TRIM_LINES:
            return text
        if not self.ansi:
            return text[:w]
        return _trim_line(text, w)

    def _trim_lines(self, lines: Lines, w: int) -> str:
        if w == 0 or not TRIM_LINES:
            return join_nl(lines)
        if not self.ansi:
            return join_nl(line[:w] for line in lines)
        return join_nl(_trim_line(line, w) for line in lines)

    ############################################################################

    def exc_lines(self, exc: BaseException, w: int = 0) -> Lines:
        self.seen.clear()
        return self._exc_lines(exc, w)

    def tb_lines(self, tb: TracebackType, w: int = 0) -> Lines:
        self.seen.clear()
        return self._tb_lines(self._tb_frames(tb), w)

    def stack_lines(self, frame: FrameType, w: int = 0) -> Lines:
        self.seen.clear()
        return self._tb_lines(self._stack_frames(frame), w)

    def task_lines(self, task: 'asyncio.Task', w: int = 0) -> Lines:
        global ASYNCIO_PATH
        if not ASYNCIO_PATH:
            import asyncio

            ASYNCIO_PATH = asyncio.__file__.removesuffix('__init__.py')

        cls = type(task)
        if cls.__name__ != 'Task':
            yield '!task'
        task_summary = f'asyncio.Task(name={task.get_name()!r}):'
        yield _BOLD(task_summary) if self.ansi else task_summary

        coro = task.get_coro()
        if isinstance(coro, CoroutineType):
            yield from self.coro_lines(coro, w, ASYNCIO_PATH)
        else:
            yield '<no coroutine stack>'

    def coro_lines(self, coro: CoroutineType, w: int = 0, avoid: str = '') -> Lines:
        if not isinstance(coro, CoroutineType):
            yield '<!coroutine>'
        self.seen.clear()
        avoid_frame = self._avoid_frame
        coro_frames = list(get_coro_stack(coro))
        frames = [f for f in coro_frames if not avoid_frame(f, avoid)]
        frames = frames or coro_frames  # avoid an empty stack
        frames.reverse()
        yield from self._tb_lines(frames, w)

    ############################################################################

    def _fmt_frame(self, frame: FrameType, width: int = 0) -> tuple[str, str]:
        code = frame.f_code
        file = code.co_filename
        line = frame.f_lineno
        if '/JetBrains/' in file:  # for scratch files
            file = file.split('/')[-1]
        loc = f'{file}:{line}'  # .removeprefix(LIB_ROOT)
        name = frame.f_code.co_qualname
        if file == '<string>':
            loc = '<generated code>'
        elif name == '<module>':
            from linecache import getline

            return loc, repr(getline(file, line).rstrip())

        args = frame_call_args(frame)
        f_name = _WHITE(name) if self.ansi else name
        key_fn = _KEY if self.ansi else str
        return loc, self.fmt.fmt_call(f_name, args, key_fn)

    def _exc_lines(self, exc: BaseException, w: int) -> Lines:
        name = type(exc).__name__
        if not isinstance(exc, BaseException):
            yield '!exception: ' + name
            return
        if isinstance(exc, RecursionError):
            yield 'RecursionError; traceback elided'
            return
        traceback = exc.__traceback__
        cause = exc.__cause__
        context = exc.__context__
        notes = getattr(exc, '__notes__', ())

        traceback = traceback if isinstance(traceback, TracebackType) else None
        cause = cause if isinstance(cause, BaseException) else None
        context = context if isinstance(context, BaseException) else None

        str_prop = self._str_lines
        strs_prop = self._strs_lines
        chain_prop = self._chain_lines

        yield from str_prop('exception', name, _EXC_CLS, w)

        if isinstance(exc, SyntaxError) and exc.filename and exc.lineno:
            location = f'{exc.filename}:{exc.lineno}'
            if text := exc.text:
                yield from str_prop('source', text, _MSG, w)
        elif traceback:
            location = frame_location(traceback_frame(traceback), False)
        else:
            location = None

        if location:
            yield from str_prop('location', location, _LOC, w)

        try:
            message = str(exc)
            assert isinstance(message, str)
        except:
            message = "<cannot format exception>"

        if message:
            yield from str_prop('message', message, _MSG, w)

        if traceback:
            tb_frames = self._tb_frames(traceback)
            yield from strs_prop('traceback', self._tb_lines(tb_frames, w), w)

        if notes:
            for i, note in enumerate(notes):
                if isinstance(note, str):
                    yield from str_prop(f'note #{i}', note, _MSG, w)

        w -= 3
        if cause:
            yield from chain_prop('cause', lambda: self._exc_lines(cause, w))  # type: ignore

        if context and context is not cause:
            yield from chain_prop('context', lambda: self._exc_lines(context, w))  # type: ignore

    def _str_lines(self, prop: str, value: str, style: FmtFn, w: int) -> Lines:
        f_prop = _PROP(prop) if self.ansi else prop
        if '\033' in value or not self.ansi:
            style = str
        value = value.strip()
        if not value:
            yield f'{f_prop}: <none>'
        if '\n' not in value:
            if len(prop) + len(value) + 3 < w and (not IN_PYCHARM or prop != 'location'):
                yield f'{f_prop}: {style(value)}'
            else:
                yield ''
                yield f'{f_prop}:'
                yield style(value)
        else:
            if prop != 'exception':
                yield ''
            yield f'{f_prop}:'
            trim = self._trim
            lines = [trim(line) for line in value.splitlines()]
            if lines and style is not str:
                maxw = min(max(map(len, lines)), w)
                for line in lines:
                    yield style(line.ljust(maxw))
            else:
                yield from lines

    def _strs_lines(self, title: str, lines: Lines, w: int) -> Lines:
        styled = _PROP(title) if self.ansi else title
        lines = list(lines)
        if len(lines) == 0:
            yield f'{styled}: <empty>'
            return
        if len(lines) == 1:
            fst = lines[0]
            if len(title) + len(fst) + 3 < w:
                yield f'{styled}: {fst}'
                return
        yield ''
        yield f'{styled}:'
        for line in lines:
            yield line

    def _chain_lines(self, title: str, line_fn: Callable[[], Lines]) -> Lines:
        title = _CHAIN(title) if self.ansi else title
        it = iter(line_fn())
        line = next(it, None)
        if line is None:
            yield f'{title}: <none>'
            return
        yield ''
        yield f'┏━━  {title} ━━━'
        yield '   ' + line
        while True:
            line = next(it, None)
            if line is None:
                break
            yield '   ' + line
        yield ''

    def _tb_lines(self, frames: Iterable[FrameType], w: int) -> Lines:
        see = self.see
        fmt_frame = self._fmt_frame
        fmt_loc = _LOC if self.ansi else str
        locs = []
        addloc = locs.append
        funs = []
        addfun = funs.append
        spill = IN_PYTEST
        for frame in frames:
            if see(frame):
                continue
            loc, fun = fmt_frame(frame)
            spill |= '\n' in fun
            addloc(loc)
            addfun(fun)
        if not locs:
            return
        locw = max(len(loc) for loc in locs)
        if spill or w and locw > w / 2:
            for loc, call in zip(reversed(locs), reversed(funs)):
                lines = list(call.splitlines())
                yield _CALL1 + lines[0]
                yield from lines[1:]
                yield _CALL2 + fmt_loc(loc)
                yield ''
        else:
            for loc, call in zip(locs, funs):
                yield _CALL1 + fmt_loc(loc.ljust(locw)) + ' ┊ ' + call

    ############################################################################

    def _stack_frames(self, frame: FrameType) -> Iterable[FrameType]:
        next_frame = frame

        frames = []
        add_frame = frames.append
        avoid_frame = self._avoid_frame

        while next_frame:
            frame = next_frame
            next_frame = frame.f_back

            if not avoid_frame(frame):
                add_frame(frame)

        frames.reverse()
        return frames

    ############################################################################

    def _tb_frames(self, tb: TracebackType | None) -> Iterable[FrameType]:
        avoid_frame = self._avoid_frame

        see = self.see
        while tb:
            frame = tb.tb_frame
            if see(tb) or avoid_frame(frame):
                pass
            else:
                yield frame
            tb = tb.tb_next

    ############################################################################

    def _avoid_frame(self, frame: FrameType, avoid_dir: str = '') -> bool:
        code = frame.f_code
        filename = code.co_filename
        qualname = code.co_qualname
        file_starts = filename.startswith
        if file_starts(PACKAGE_DIRS):
            rel_path = package_relative_path(filename)
            if rel_path.startswith(SYS_IGNORED_PACKAGES):
                return True
        if file_starts(SYS_IGNORED_FILENAMES):
            return True

        if code.co_name.startswith('raise_'):
            return True
        if qualname.startswith(SYS_IGNORED_QUALNAMES):
            return True
        if avoid_dir and file_starts(avoid_dir):
            return True
        if '<genexpr>' in code.co_qualname:
            return True

        starts = filename.startswith
        if not starts(self.root_path):
            return False
        if any(starts(a) for a in self.avoided_files):
            return True

        avoid_func = self.avoided_funcs.__contains__
        if avoid_func(qualname):
            return True
        if '.<locals>.' in qualname:
            if avoid_func(qualname.split('.<locals>.', 1)[0]):
                return True

        fn = frame_function(frame)
        if avoid_func(id(fn)):
            return True
        if isinstance(fn, MethodType) and avoid_func(id(fn.__func__)):
            return True
        return False


################################################################################

_CODE_RE = re.compile('(\033' + r'\[[0-?]+m)')


def _trim_line(text: str, w: int) -> str:
    if not TRIM_LINES:
        return text
    if w == 0 or len(text) < w:
        return text

    chunks = []
    add = chunks.append
    right = text
    needs_reset = False
    if '\033' not in text:
        return text[:w]
    rem = w
    while True:
        split: list[str] = _CODE_RE.split(right, 1)
        if len(split) == 1:
            add(right[:rem])
            break
        elif len(split) == 3:
            left, code, right = split
        else:
            break
        assert isinstance(left, str)
        rem -= len(left)
        if rem < 0:
            add(left[:rem])
            break
        add(left)
        if rem == 0:
            break
        needs_reset = code != _RESET
        add(code)
    if needs_reset:
        add(_RESET)
    return ''.join(chunks)


def p_int(obj: int) -> str:
    if I48 <= abs(obj) <= I64:
        return f'0x{obj:016x}'
    if I24 <= obj < I48:
        return f'0x{obj:010x}'
    return str(obj)


def p_strlike(obj: str | bytes | bytearray) -> str:
    size = len(obj)
    if size <= 24:
        return repr(obj)
    name = type(obj).__name__
    short = repr(obj[:22])
    short = f'{short[:-2]}…{size - 22}{short[-1]}'
    return f'<{name} {short}>'


def p_function(func: Callable | str) -> str:
    if isinstance(func, str):
        return func
    if hasattr(func, '__qualname__'):
        return func.__qualname__
    elif hasattr(func, '__name__'):
        return func.__name__
    return '<callable>'


def p_method(meth: MethodType) -> str:
    self = meth.__self__
    func = meth.__func__
    self_name = p_type(type(self)) + '.'
    func_name = getattr(func, '__qualname__', '')
    p_self = p_object(self)
    if func_name.startswith(self_name):
        return func_name
    p_func = p_function(func)
    return f'{p_self}.{p_func}'


def p_type(cls: type) -> str:
    if not is_virt_cls(cls) and hasattr(cls, '__short_name__'):
        return cls.__short_name__
    return cls.__qualname__


def p_object(obj: object, extra: str = '') -> str:
    name = p_type(type(obj))
    ptr = id(obj)
    if extra:
        name = name + ' ' + extra
    else:
        info = getattr(obj, '__debug_info_str__', None)
        if callable(info):
            try:
                info_str = info()
                if isinstance(info_str, str):
                    name = name + ' ' + info_str
            except Exception:
                pass
    return f"<{name} @ {ptr:010x}>"


def a_int(obj: int) -> str:
    if I48 <= abs(obj) <= I64:
        return _rgb_hash(obj, f'0x{obj:016x}')
    if I24 <= obj < I48:
        return _rgb_ptr(obj, f'0x{obj:010x}')
    return str(obj)


def a_function(func: Callable | str) -> str:
    return _FUN(p_function(func))


def a_method(meth: MethodType) -> str:
    self = meth.__self__
    func = meth.__func__
    self_name = p_type(type(self)) + '.'
    func_name = getattr(func, '__qualname__', '')
    a_self = a_object(self)
    if func_name.startswith(self_name):
        a_func = a_function(func_name.removeprefix(self_name))
        return f'{a_self}.{a_func}'
    a_func = a_function(func)
    return f'{a_self}.{a_func}'


def a_type(cls: type) -> str:
    if hasattr(cls, '__short_name__'):
        return _CLS(cls.__short_name__)
    return _CLS(cls.__qualname__)


def a_object(obj: object, extra: str = '') -> str:
    cls = type(obj)
    name = p_type(cls)
    if extra:
        name = name + ' ' + extra
    if cls is type or isinstance(obj, type):
        if is_virt_cls(obj):
            name = 'virtual ' + name + f" '{obj.__name__}"
    elif is_virt_obj(obj):
        name = 'virtual ' + name
    elif callable(info := getattr(cls, '__debug_info_str__', None)):
        try:
            if isinstance(res := info(obj), str) and res:
                name = name + ' ' + res
        except Exception:
            pass
    return _rgb_ptr(id(obj), f'<{name}>')


def is_virt_cls(cls: type) -> bool:
    try:
        cls_get(cls, '___vhdl___')
        return True
    except:
        return False


def is_virt_obj(obj: object) -> bool:
    try:
        obj_get(obj, '___vhdl___')
        return True
    except:
        return False


obj_get = object.__getattribute__
cls_get = type.__getattribute__


def _rgb_ptr(ptr: int, s: str) -> str:
    return _rgb_hash(raw_int_hash(ptr), s)


def _rgb_hash(h: int, s: str) -> str:
    r, g, b = (abs(h) & 0xFFFFFF).to_bytes(3)
    return _RGB(r, g, b, s)


COMMA = ', '.join


I24 = 1 << 24
I48 = 1 << 48
I64 = 1 << 64
M57 = (1 << 57) - 1


def raw_int_hash(i: int) -> int:
    n = 0xCBF29CE484222325 ^ i
    n = n ^ (n >> 33)
    n = n * 0xFF51AFD7ED558CCD
    n = n ^ (n >> 33)
    n = n * 0xC4CEB9FE1A85EC53
    n = n ^ (n >> 33)
    return n & M57


# so it doesnt trigger before user fns
class _object: ...


PLAIN_DISPATCH: dict[type, FmtFn] = {
    NoneType:            str,
    bool:                str,
    int:                 p_int,
    str:                 p_strlike,
    bytes:               p_strlike,
    bytearray:           p_strlike,
    FunctionType:        p_function,
    BuiltinFunctionType: p_function,
    MethodType:          p_method,
    type:                p_type,
    _object:             p_object,
}

ANSI_DISPATCH: dict[type, FmtFn] = PLAIN_DISPATCH | {
    int:                 a_int,
    FunctionType:        a_function,
    BuiltinFunctionType: a_function,
    MethodType:          a_method,
    type:                a_type,
    _object:             a_object,
}

join_nl = '\n'.join

_RESET   = '\033[0m'
_RGB     = '\033[38;2;{};{};{}m{}\033[39m'.format

def _ANSI(code: str) -> Callable[..., str]:
    return f'\033[{code}m{{}}\033[0m'.format  # type: ignore

_WARN    = _ANSI('1;43;37')
_ERROR   = _ANSI('1;41;37')
_WHITE   = _ANSI('37')
_RED     = _ANSI('31')
_GREEN   = _ANSI('32')
_YELLOW  = _ANSI('33')
_BLUE    = _ANSI('34')
_MAGENTA = _ANSI('35')
_CYAN    = _ANSI('36')
_BOLD    = _ANSI('1')
_DIM     = _ANSI('2')
_ITALIC  = _ANSI('3')
_UNDER   = _ANSI('4')

_SLINE   = '═' * 50
_PLINE   = '─' * 50
_ALINE   = _DIM(_PLINE)

_PTICK  = '┃  '
_ATICK  = _DIM(_PTICK)
# _PTICKS  = '┌  ', '│  '
# _ATICKS  = _DIM('┌  '), _DIM('│  ')
# colors of actual things
_FUN     = _YELLOW
_PROP    = _ANSI('3;32')
_LOC     = _ANSI('3;90')
_CLS     = _CYAN

_MSG     = _ANSI('1;100')

_KEY     = _ANSI('3;94')
_EXC_SYS = _ANSI('45')
_EXC_CLS = _ANSI('1;35')
_CHAIN   = _ANSI('3;34')

_CALL1   = '' if IN_PYCHARM else '→ '
_CALL2   = '' if IN_PYCHARM else '  '


def term_width() -> int:
    global TERM_WIDTH
    if TERM_WIDTH:
        return TERM_WIDTH
    try:
        TERM_WIDTH = 80
        env_cols = environ.get('COLUMNS', '')
        if env_cols and all(s.isdigit() for s in env_cols.split()):
            TERM_WIDTH = int(env_cols)
        elif IN_PYCHARM:
            TERM_WIDTH = 100
        else:
            TERM_WIDTH = shutil.get_terminal_size().columns
    except Exception:
        pass
    return TERM_WIDTH


def justified(title: str, bg: str, width: int = 0) -> str:
    width = width or TERM_WIDTH or term_width()
    h = (len(bg) - 1) // 2
    l, r = bg[:h], bg[h + 1 :]
    m: str = bg[h]
    line = l + (m * (width - len(l) - len(r))) + r
    if not title:
        return line
    title = ' ' + title + ' '
    x = width // 2 - len(title) // 2
    return line[:x] + title + line[x + len(title) :]


class ColorizeCache(dict[str, str]):
    def __missing__(self, key: str) -> str:
        col = Rgb.hash_str(key)
        res = col.hi(key)
        if len(key) <= 20:
            self[key] = res
        return res


_CCACHE = ColorizeCache()


def colorize(s: str, n: int = 0) -> str:
    if not isinstance(s, str):
        s = str(s)
    if len(s) > 64 or '\033' in s or '\n' in s:
        return s
    s = _CCACHE[s]
    return s + ' ' * (n - len(s)) if n else s


def get_coro_stack(coro: CoroutineType) -> Iterable[FrameType]:
    while type(coro) is CoroutineType:
        if cr_frame := coro.cr_frame:
            while cr_frame:
                yield cr_frame
                cr_frame = cr_frame.f_back
        coro = coro.cr_await


def _package_dirs() -> tuple[str, ...]:
    try:
        import site
        import sysconfig

        packages = site.getsitepackages()
        packages.append(sysconfig.get_path('stdlib'))
        return tuple(abspath(p) for p in packages)
    except:
        return ()


def package_relative_path(filename: str) -> str:
    starts = filename.startswith
    for p_dir in PACKAGE_DIRS:
        if starts(p_dir):
            rel = filename.removeprefix(p_dir).removeprefix('/')
            if 'site-packages' not in rel:
                return rel.replace('/', '.').removesuffix('.py')
    return ''


####################################################################################################

PANIC_BANNER = '''
███████████████
██   PANIC   ██
███████████████
'''

####################################################################################################

PACKAGE_DIRS = _package_dirs()
SYS_IGNORED_PACKAGES = ('pytest', '_pytest', 'pluggy', 'pytest_asyncio', 'asyncio.runners')
SYS_IGNORED_FILENAMES = ('<frozen',)
SYS_IGNORED_QUALNAMES = (
    'Thread._bootstrap',
    'BaseEventLoop.run_until_complete',
)
ASYNCIO_PATH = ''

ROOT_PATH = Path(__file__).parent.parent.as_posix()

####################################################################################################

debug_fmt = DebugFormatter(root_path=ROOT_PATH)

fmt_frame = debug_fmt.fmt_frame
fmt_exception = debug_fmt.fmt_exception
fmt_traceback = debug_fmt.fmt_traceback
fmt_current_stack = debug_fmt.fmt_current_stack

avoid_file = debug_fmt.avoid_file
avoid_function = debug_fmt.avoid_func
avoid_module = debug_fmt.avoid_module

enable_rich_tracebacks = debug_fmt.use_for_uncaught_exceptions

####################################################################################################

NO_REC_LIMIT = 1000



# ruff: noqa
# fmt: off

"""
This a hand-curated typeshed file for the builtin `types` module.
"""

from types import ModuleType
from typing import (
    Any,
    Callable,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    NoReturn,
    Self,
    ValuesView,
    overload,
)

################################################################################

__all__ = [
    'FunctionType',
    'LambdaType',
    'CodeType',
    'CellType',
    'BuiltinFunctionType',
    'BuiltinMethodType',
    'MethodType',
    'MethodWrapperType',
    'MappingProxyType',
    'GetSetDescriptorType',
    'MemberDescriptorType',
    'MethodDescriptorType',
    'ClassMethodDescriptorType',
    'WrapperDescriptorType',
    'EllipsisType',
    'NoneType',
    'NotImplementedType',
    'GeneratorType',
    'AsyncGeneratorType',
    'CoroutineType',
    'TracebackType',
    'FrameType',
    'ModuleType',
    'UnionType',
    'GenericAlias',
    'SimpleNamespace',
]

################################################################################

# for internal use...

type Objs = tuple[object, ...]
type Strs = tuple[str, ...]
type OptStr = str | None
type OptObjs = Objs | None
type OptDict = dict[str, Any] | None
type AnyDict = dict[str, Any]
type TypeParams = tuple[Any, ...]
type OptCells = tuple[CellType, ...] | None
type OptTraceback = TracebackType | None


def _obj_name(obj: object) -> str:
    cls_name = _cls_name(obj)
    ptr_name = _ptr_name(obj)
    return f"{cls_name} object at {ptr_name}"


def _cls_name(obj: object) -> str:
    return obj.__class__.__name__


def _ptr_name(obj: object) -> str:
    return f'0x{id(obj):10x}'


################################################################################


class PyTypeMeta(type):
    __print_form__: str = ''  # what repr() will print for this type
    __alias_name__: str = ''  # what does core.type.sys call this?
    __source_loc__: bytes = b''  # which file is the PyTypeObject defined in, and what is it called?


class PyType(metaclass=PyTypeMeta):
    """
    Represents a CPython-implemented type. These are all immutable.
    """


################################################################################


class CanSort:
    def __lt__(self, other: object) -> bool: ...
    def __le__(self, other: object) -> bool: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...


class CanEqual:
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, other: object) -> bool: ...


class CanHash:
    def __hash__(self) -> int: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, other: object) -> bool: ...


class CanCall:
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...


class HasSig:
    __text_signature__: OptStr


class HasModule:
    __doc__: OptStr


class HasName:
    __name__:     str
    __qualname__: str
    __doc__:      OptStr


class HasDerivedName:
    @property
    def __doc__(self) -> OptStr: ...
    @property
    def __name__(self) -> str: ...
    @property
    def __qualname__(self) -> str: ...


class BindsSelf:
    __self__: object


class BindsClass:
    __objclass__: type


################################################################################


class CellType(PyType, CanEqual):
    __print_form__ = 'cell'
    __alias_name__ = 'CellT'
    __source_loc__ = b'Objects/cellobject.c#PyCell_Type'

    __hash__ = None

    cell_contents: Any

    def __new__(cls, contents: object = ..., /) -> Self: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, value: object, /) -> bool: ...

    # emulate what str() would do
    def __str__(self) -> str:
        ptr_name = _ptr_name(self)
        val_name = _obj_name(self.cell_contents)
        return f"<cell at {ptr_name}: {val_name}>"


class CodeType(PyType, CanCall, CanHash):
    __print_form__ = 'code'
    __alias_name__ = 'CodeT'
    __source_loc__ = b'Objects/codeobject.c#PyCode_Type'

    # methods:
    def __new__(
        cls,
        argcount: int,
        posonlyargcount: int,
        kwonlyargcount: int,
        nlocals: int,
        stacksize: int,
        flags: int,
        codestring: bytes,
        constants: Objs,
        names: Strs,
        varnames: Strs,
        filename: str,
        name: str,
        qualname: str,
        firstlineno: int,
        linetable: bytes,
        exceptiontable: bytes,
        freevars: Strs = ...,
        cellvars: Strs = ...,
        /,
    ) -> Self: ...

    def replace(
        self,
        *,
        co_argcount: int = -1,
        co_posonlyargcount: int = -1,
        co_kwonlyargcount: int = -1,
        co_nlocals: int = -1,
        co_stacksize: int = -1,
        co_flags: int = -1,
        co_firstlineno: int = -1,
        co_code: bytes = ...,
        co_consts: Objs = ...,
        co_names: Strs = ...,
        co_varnames: Strs = ...,
        co_freevars: Strs = ...,
        co_cellvars: Strs = ...,
        co_filename: str = ...,
        co_name: str = ...,
        co_qualname: str = ...,
        co_linetable: bytes = ...,
        co_exceptiontable: bytes = ...,
    ) -> Self: ...

    def __sizeof__(self) -> int: ...
    def __co_lines__(self): ...
    def co_branches(self): ...
    def co_positions(self): ...
    def _varname_from_oparg(self): ...

    __replace__ = replace

    # members:
    co_argcount: int
    co_posonlyargcount: int
    co_kwonlyargcount: int
    co_stacksize: int
    co_flags: int
    co_nlocals: int
    co_consts: Objs
    co_names: Strs
    co_filename: str
    co_name: str
    co_qualname: str
    co_firstlineno: int
    co_linetable: bytes
    co_exceptiontable: bytes
    co_lnotab: bytes
    _co_code_adaptive: bool

    # marked as "backward compat" in CPython source, whatever that means:
    co_varnames: Strs
    co_cellvars: Strs
    co_freevars: Strs
    co_code: bytes

    # from CanCall
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

    # from CanHash
    def __hash__(self) -> int: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

    # emulate what str() would do
    def __str__(self) -> str:
        code_name = self.co_name
        ptr_name = _ptr_name(self)
        file_name = repr(self.co_filename)
        line_name = self.co_firstlineno
        return f"<code object {code_name} at {ptr_name}, file {file_name}, line {line_name}>"


################################################################################
# inspect._signature_fromstr


class IsFunction(CanCall, HasName, HasModule, CanHash):
    pass


class FunctionType(PyType, IsFunction):
    __print_form__ = 'function'
    __alias_name__ = 'FunctionT'
    __source_loc__ = b'Objects/funcobject.c#FunctionTion_Type'

    def __new__(
        cls,
        code: CodeType,
        globals: AnyDict,
        name: OptStr = None,
        argdefs: Objs | None = None,
        closure: OptCells = None,
        kwdefaults: OptDict = None,
    ): ...

    __closure__:     OptCells
    __globals__:     AnyDict
    __module__:      str      # from HasModule
    __builtins__:    AnyDict
    __code__:        CodeType
    __defaults__:    OptObjs
    __kwdefaults__:  OptDict
    __dict__:        dict
    __annotations__: AnyDict  # type: ignore
    __type_params__: TypeParams
    __name__:        str      # from HasName
    __qualname__:    str      # from HasName
    __doc__:         OptStr   # from HasName

    # from CanCall
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

    # from CanHash
    def __hash__(self) -> int: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

    # emulate what str() would do
    def __str__(self) -> str:
        func_name = self.__name__
        ptr_name = _ptr_name(self)
        return f"<function {func_name} at {ptr_name}>"


LambdaType = FunctionType


class BuiltinFunctionType(PyType, IsFunction, BindsSelf, HasSig):
    __print_form__ = 'builtin_function_or_method'
    __alias_name__ = 'BoundMethodOrFuncC'
    __source_loc__ = b'Objects/methodobject.c#PyCFunction_Type'

    __self__:           object | ModuleType  # from IsBound
    __name__:           str                  # from HasName
    __qualname__:       str                  # from HasName
    __doc__:            str                  # from HasName
    __module__:         str                  # from HasModule
    __text_signature__: OptStr               # from HasSig

    # NOTE:
    # `__self__ = builtins` for  `len`
    # `__self__ = int` for `int.__new__`

    # from CanCall
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

    # from CanHash
    def __hash__(self) -> int: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

    # emulate what str() would do
    def __str__(self) -> str:
        func_name = self.__name__
        if isinstance(self.__self__, ModuleType):
            return f"<built-in function {func_name}>"
        self_name = _obj_name(self.__self__)
        return f"<built-in method {func_name} of {self_name}>"


################################################################################


class IsMethod(CanCall, HasDerivedName, CanHash):
    pass


class MethodType(PyType, IsMethod, BindsSelf):
    """Stores a python-implemented INSTANCE method and the OBJECT it is bound to.

    Example:
        `my_object.my_instance_method`
    """

    __print_form__ = 'method'
    __alias_name__ = 'BoundMethodT'
    __source_loc__ = b'Objects/classobject.c#PyInstanceMethod_Type'

    __self__: object         # from BindsSelf
    __func__: FunctionType

    def __new__(cls, fn: Callable): ...

    # obtained from inner function
    @property
    def __closure__(self) -> OptCells: ...
    @property
    def __defaults__(self) -> OptObjs: ...

    # from CanCall
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

    # from HasDerivedName
    @property
    def __doc__(self) -> OptStr: ...
    @property
    def __name__(self) -> str: ...
    @property
    def __qualname__(self) -> str: ...

    # from CanHash
    def __hash__(self) -> int: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

    # emulate what str() would do
    def __str__(self) -> str:
        meth_name = self.__func__.__name__
        self_name = _obj_name(self.__self__)
        return f"<bound method {meth_name} of <{self_name}>>"


class MethodWrapperType(PyType, IsMethod, BindsSelf, HasSig):
    """Stores the name of a POLYMORPHIC OPERATION and the OBJECT it is bound
    to. The partially-bound version of this (bound to class, but not instance)
    is a WrapperDescriptorType.

    Explanation:
        POLYMORPHIC operations corresponding to built functions `len`, `hash`,
        `str`, `repr`, and `iter`, have single C implementations that do their
        own dispatch. This type is used to store the intended operation AND
        the object it will operate on.

        Bound MONOMORPHIC instance methods often use `BuiltinFunctionType`, which
        can also represent bound methods, e.g. `[].append`.

    Example:
        `[].__len__`
        `().__str__`

    Note:
        MONOMORPHIC instance methods will use `BuiltinFunctionType`, which
        can also represent bound methods, e.g. `[].append`.
    """

    __print_form__ = 'method-wrapper'
    __alias_name__ = 'BoundDunderMethodC'
    __source_loc__ = b'Objects/descrobject.c#_PyMethodWrapper_Type'

    __self__: object            # from BindsSelf
    __text_signature__: OptStr  # from HasSig

    # derived from __self__
    def __objclass__(self) -> type:
        return self.__self__.__class__

    # from CanCall
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

    # from HasDerivedName
    @property
    def __doc__(self) -> OptStr: ...
    @property
    def __name__(self) -> str: ...
    @property
    def __qualname__(self) -> str: ...

    # from CanHash
    def __hash__(self) -> int: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

    # emulate what str() would do
    def __str__(self) -> str:
        meth_name = self.__name__
        self_name = _obj_name(self.__self__)
        return f"<method-wrapper '{meth_name}' of {self_name}>"


BuiltinMethodType = BuiltinFunctionType


################################################################################


class IsDescriptor(HasDerivedName, BindsClass, CanHash):
    """Abstract class representing read-only descriptor types."""

    __objclass__: type  # from BindsClass

    def __get__(self, instance: Any, owner: type | None = None, /) -> Any: ...


class IsMutableDescriptor(IsDescriptor):
    """Abstract class representing read-write descriptor types."""

    def __get__(self, instance: Any, owner: type | None = None, /) -> Any: ...
    def __set__(self, instance: Any, value: Any, /) -> None: ...
    def __delete__(self, instance: Any, /) -> None: ...


################################################################################


# properties implemented in C via PyGetSetDef structure, e.g. complex.real, frame.f_locals
class GetSetDescriptorType(PyType, IsMutableDescriptor):
    """Stores the name of an DESCRIPTOR implementation and the CLASS it is for,
    which intercepts get, set, and delete operations on objects of that class
    when the attribute (this thing) is read, written, or changed via the
    owning object.
    """

    __print_form__ = 'getset_descriptor'
    __alias_name__ = 'MutablePropertyC'
    __source_loc__ = b'Objects/object.c#PyGetSetDescr_Type'

    __objclass__: type  # from BindsClass

    # from IsMutableDescriptor
    def __get__(self, instance: Any, owner: type | None = None, /) -> Any: ...
    def __set__(self, instance: Any, value: Any, /) -> None: ...
    def __delete__(self, instance: Any, /) -> None: ...

    # from HasDerivedName
    @property
    def __doc__(self) -> OptStr: ...
    @property
    def __name__(self) -> str: ...
    @property
    def __qualname__(self) -> str: ...

    # from CanHash
    def __hash__(self) -> int: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

    # emulate what str() would do
    def __str__(self) -> str:
        attr_name = self.__name__
        cls_name = self.__objclass__.__name__
        return f"<attribute '{attr_name}' of '{cls_name}' objects>"


# direct struct member access in C via PyMemberDef, e.g. __weakref__, code.co_name
# faster, but less flexible. Handles slots.
class MemberDescriptorType(PyType, IsMutableDescriptor):
    """Stores the name of a SLOT and the CLASS it is a slot for. It can't be
    called, it is there to satisfy the descriptor protocol to enable slot
    setting and getting to work.

    Example:
        ```
        class HasSlots:
            __slots__ = ('a',)

        HasSlots.a
    """

    __print_form__ = 'slot'
    __alias_name__ = 'SlotPropertyC'
    __source_loc__ = b'Objects/object.c#PyMemberDescr_Type'

    __objclass__: type  # from BindsClass

    # from IsMutableDescriptor
    def __get__(self, instance: Any, owner: type | None = None, /) -> Any: ...
    def __set__(self, instance: Any, value: Any, /) -> None: ...
    def __delete__(self, instance: Any, /) -> None: ...

    # from HasDerivedName
    @property
    def __doc__(self) -> OptStr: ...
    @property
    def __name__(self) -> str: ...
    @property
    def __qualname__(self) -> str: ...

    # from CanHash
    def __hash__(self) -> int: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

    # emulate what str() would do
    def __str__(self) -> str:
        attr_name = self.__name__
        cls_name = self.__objclass__.__name__
        return f"<slot '{attr_name}' of '{cls_name}' objects>"


################################################################################


class IsMethodDescriptor(IsDescriptor, CanCall, HasSig):
    """Abstract class representing callable descriptor types."""


################################################################################


# bytes.hex, list.append, dict.update
class MethodDescriptorType(PyType, IsMethodDescriptor):
    """Stores the name of C implementation of an INSTANCE method
    and the CLASS it is bound to, and awaits a `self` and arguments.

    Example: `bytes.hex`, where `__objcls__ = hex`.

    The bound equivalent of this is `BuiltinFunctionType` aka `BoundMethodOrFuncC`.
    """

    __print_form__ = 'method_descriptor'
    __alias_name__ = 'UnboundMethodC'
    __source_loc__ = b'Objects/descrobject.c#PyMethodDescr_Type'

    __objclass__: type          # from BindsClass
    __text_signature__: OptStr  # from HasSig

    # from CanCall
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

    # from IsDescriptor
    def __get__(self, instance: Any, owner: type | None = None, /) -> Any: ...

    # from HasDerivedName
    @property
    def __doc__(self) -> OptStr: ...
    @property
    def __name__(self) -> str: ...
    @property
    def __qualname__(self) -> str: ...

    # from CanHash
    def __hash__(self) -> int: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

    # emulate what str() would do
    def __str__(self) -> str:
        meth_name = self.__name__
        cls_name = self.__objclass__.__name__
        return f"<method '{meth_name}' of '{cls_name}' objects>"


class ClassMethodDescriptorType(IsMethodDescriptor):
    """Stores the name of a C implementation of an CLASS method AND
    the CLASS it is bound to, and awaits just arguments.

    Example: `bytes.fromhex`, where `__objcls__ = hex`

    There is no unbound version of this!
    """

    __print_form__ = 'classmethod_descriptor'
    __alias_name__ = 'BoundClassMethodC'
    __source_loc__ = b'Objects/descrobject.c#PyClassMethodDescr_Type'

    __objclass__: type          # from BindsClass
    __text_signature__: OptStr  # from HasSig

    # from CanCall
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

    # from IsDescriptor
    def __get__(self, instance: Any, owner: type | None = None, /) -> Any: ...

    # from HasDerivedName
    @property
    def __doc__(self) -> OptStr: ...
    @property
    def __name__(self) -> str: ...
    @property
    def __qualname__(self) -> str: ...

    # from CanHash
    def __hash__(self) -> int: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

    # emulate what str() would do
    def __str__(self) -> str:
        meth_name = self.__name__
        cls_name = self.__objclass__.__name__
        return f"<method '{meth_name}' of '{cls_name}' objects>"


class WrapperDescriptorType(IsMethodDescriptor):
    """Stores the name of a POLYMORPHIC OPERATION, and the CLASS it is bound to.
    This awaits an INSTANCE and arguments.

    The fully-bound version of this (bound to a concrete instance rather than
    just the class) is a MethodWrapperType.

    Examples:
        `bytes.__buffer__`
        `object.__init__`
        `object.__lt__`
        `type.__init__`
    """

    __print_form__ = 'wrapper_descriptor'
    __alias_name__ = 'UnboundDunderMethodC'
    __source_loc__ = b'Objects/descrobject.c#PyWrapperDescr_Type'

    __objclass__: type          # from BindsClass
    __text_signature__: OptStr  # from HasSig

    # from CanCall
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

    # from IsDescriptor
    def __get__(self, instance: Any, owner: type | None = None, /) -> Any: ...

    # from HasDerivedName
    @property
    def __doc__(self) -> OptStr: ...
    @property
    def __name__(self) -> str: ...
    @property
    def __qualname__(self) -> str: ...

    # from CanHash
    def __hash__(self) -> int: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

    # emulate what str() would do
    def __str__(self) -> str:
        slot_func_name = self.__name__
        cls_name = self.__objclass__.__name__
        return f"<slot wrapper '{slot_func_name}' of '{cls_name}' objects>"


################################################################################


# not a real superclass
class PySingleton(PyType):
    def __init__(self) -> None: ...


class EllipsisType(PySingleton):
    __print_form__ = 'ellipsis'
    __alias_name__ = 'EllipT'
    __source_loc__ = b'Objects/sliceobject.c#PyEllipsis_Type'

    # emulate what str() would do
    def __str__(self) -> str:
        return '...'


class NoneType(PySingleton):
    __print_form__ = 'NoneType'
    __alias_name__ = 'NoneT'
    __source_loc__ = b'Objects/object.c#_PyNone_Type'

    # emulate what str() would do
    def __str__(self) -> str:
        return 'None'


class NotImplementedType(PySingleton):
    __print_form__ = 'NotImplementedType'
    __alias_name__ = 'NotImplT'
    __source_loc__ = b'Objects/object.c#_PyNotImplemented_Type'

    # emulate what str() would do
    def __str__(self) -> str:
        return 'NotImplemented'


################################################################################


class FrameType(PyType):
    __print_form__ = 'frame'
    __alias_name__ = 'FrameT'
    __source_loc__ = b'Objects/frameobject.c#PyFrame_Type'

    # methods
    def __sizeof__(self) -> int: ...
    def clear(self) -> None: ...

    # members
    f_back: Self | None
    f_locals: AnyDict
    f_lineno: int | Any  # writeable
    f_trace: Callable[[Self, str, Any], Any] | None  # writeable
    f_lasti: int
    f_globals: AnyDict
    f_builtins: AnyDict
    f_code: CodeType
    f_trace_opcodes: bool  # writeable
    f_generator: 'GeneratorType | None'

    # getset:
    f_trace_lines: bool


class TracebackType(PyType):
    __print_form__ = 'traceback'
    __alias_name__ = 'TracebackType'
    __source_loc__ = b'Python/Traceback.c#PyTraceBack_Type'

    # methods:
    def __new__(
        cls, tb_next: Self | None, tb_frame: FrameType, tb_lasti: int, tb_lineno: int
    ) -> Self: ...
    def __dir__(self) -> list[str]: ...

    # members:
    tb_frame: FrameType
    tb_lasti: int

    # getset:
    tb_next: Self | None  # writeable
    tb_lineno: int


################################################################################


class CoroutineType(PyType):
    __print_form__ = 'coroutine'
    __alias_name__ = 'CoroutineT'
    __source_loc__ = b'Objects/genobject.c#PyCoro_Type'

    # methods:
    def send(self, arg: Any, /) -> Any: ...
    async def throw(self, typ: BaseException) -> Any: ...

    def close(self) -> NoReturn: ...
    def __sizeof__(self) -> int: ...
    def __class_getitem__(cls, item: Any, /) -> Self: ...

    # getset:
    __name__: str
    __qualname__: str
    cr_running: bool
    cr_await: object | None  # Awaitable or None
    cr_frame: FrameType | None
    cr_suspended: bool
    cr_origin: (
        tuple[tuple[str, int, str], ...] | None
    )  # depends on sys.set_coroutine_origin_tracking_depth


class GeneratorType(PyType):
    __print_form__ = 'generator'
    __alias_name__ = 'GeneratorT'
    __source_loc__ = b'Objects/genobject.c#PyGen_Type'

    # methods:
    def __iter__(self) -> Self: ...
    def __next__(self) -> Any: ...
    def send(self, arg: Any, /) -> Any: ...
    def throw(self, typ: BaseException, /) -> Any: ...
    def close(self) -> NoReturn: ...
    def __sizeof__(self) -> int: ...
    def __class_getitem__(cls, item: Any, /) -> Any: ...

    # getset:
    __name__: str
    __qualname__: str
    gi_yieldfrom: Self | None
    gi_running: bool
    gi_frame: FrameType | None
    gi_suspended: bool
    gi_code: CodeType


class AsyncGeneratorType(PyType):
    __print_form__ = 'async_generator'
    __alias_name__ = 'AGeneratorT'
    __source_loc__ = b'Objects/genobject.c#PyAsyncGen_Type'

    # methods:
    def __aiter__(self) -> Self: ...
    def __anext__(self) -> CoroutineType: ...
    def asend(self, arg: Any, /) -> Any: ...
    async def athrow(self, val: BaseException) -> Any: ...
    def aclose(self) -> NoReturn: ...
    def __sizeof__(self) -> int: ...
    def __class_getitem__(cls, item: Any, /) -> CoroutineType: ...

    # getset:
    __name__: str
    __qualname__: str
    ag_running: bool
    ag_await: object | None  # Awaitable or None
    ag_frame: FrameType | None
    ag_suspended: bool
    ag_code: CodeType


################################################################################


class ModuleSpec:
    """
    Proxy for class implementing importlib.machinery.ModuleSpec, there is no CPython type for it.
    Example implementing type is `_frozen_importlib.ModuleSpec`.
    """

    name: str
    loader: Any
    origin: Any
    submodule_search_locations: list[str]
    loader_state: Any
    cached: OptStr
    parent: str
    has_location: bool


class ModuleLoader:
    """
    Proxy for class implementing importlib.abc.Module, there is no CPython type for it.
    Example implementing type is `_frozen_importlib.ModuleSpec`.
    """

    def create_module(self, spec: ModuleSpec) -> 'ModuleType | None': ...
    def exec_module(self, module: 'ModuleType') -> None: ...


class ModuleType(PyType):
    __print_form__ = 'module'
    __alias_name__ = 'ModuleT'
    __source_loc__ = b'Objects/moduleobject.c#PyFrame_Type'

    # methods:
    def __init__(self, name: str, doc: str | None = ...) -> None: ...
    def __dir__(self) -> dict[str, Any]: ...

    # these are dynamically created and placed in `__dict__`:
    __name__:       str
    __file__:       OptStr     # not necessarily present
    __path__:       list[str]  # not necessarily present
    __package__:    OptStr
    __spec__:       ModuleSpec | None
    __loader__:     ModuleLoader | None
    __doc__:        OptStr
    __cached__:     OptStr     # deprecated, will be removed in 3.15
    __builtins__:   AnyDict  # not necessarily present
    # this has getter/setter but modifies __dict__['__annotations__']
    __annotations_: AnyDict


################################################################################


class UnionType(PyType):
    __print_form__ = 'module'
    __alias_name__ = 'ModuleT'
    __source_loc__ = b'Objects/moduleobject.c#PyModule_Type'

    __args__: Objs

    def __or__(self, value: Any, /) -> Self: ...
    def __ror__(self, value: Any, /) -> Self: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __hash__(self) -> int: ...


################################################################################


class GenericAlias(PyType):
    __print_form__ = 'types.GenericAlias'
    __alias_name__ = 'CPythonGenericAlias'
    __source_loc__ = b'Objects/genericaliasobject.c#PyModule_Type'

    # methods:
    def __new__(cls, origin: type, args: Any) -> Self: ...
    def __getitem__(self, typeargs: Any, /) -> Self: ...
    def __getattr__(self, name: str) -> Any: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __hash__(self) -> int: ...
    def __or__(self, value: Any, /) -> UnionType: ...
    def __ror__(self, value: Any, /) -> UnionType: ...

    # members (readonly):
    __origin__: type
    __args__: Objs
    __unpacked__: bool

    # getset (readonly):
    __parameters__: TypeParams
    __typing_unpacked_tuple_args__: OptObjs


################################################################################


class SimpleNamespace(PyType):
    def __init__(
        self, mapping_or_iterable: Iterable[tuple[str, object]] = (), /, **kwargs: object
    ): ...

    def __repr__(self) -> str: ...
    def __eq__(self, other) -> bool: ...

    def __replace__(self, /, **changes: object) -> None: ...


################################################################################


class MappingProxyType(PyType, CanSort):
    __print_form__ = 'mappingproxy'
    __alias_name__ = 'MapProxyT'
    __source_loc__ = b'Objects/descrobject.c#PyDictProxy_Type'

    __hash__ = None

    # methods:
    def __new__(cls, mapping: Any) -> Self: ...
    def __getitem__(self, key: Any, /) -> Any: ...
    def __iter__(self) -> Iterator[Any]: ...
    def __len__(self) -> int: ...
    def __class_getitem__(cls, item: Any, /) -> GenericAlias: ...
    def __reversed__(self) -> Iterator[Any]: ...
    def __or__(self, value: Any, /) -> dict: ...
    def __ror__(self, value: Any, /) -> dict: ...
    def copy(self) -> dict: ...
    def keys(self) -> KeysView[Any]: ...
    def values(self) -> ValuesView[Any]: ...
    def items(self) -> ItemsView[Any, Any]: ...
    @overload
    def get(self, key: Any, /) -> Any | None: ...
    @overload
    def get(self, key: Any, default: Any, /) -> Any: ...
    def get(self, key: Any, default: Any = None, /) -> Any: ...



# fmt: off

from .__ import *
from .base import Msg


__all__ = [
    'RequestMsg',
]

################################################################################

if TYPE_CHECKING:
    from ..request.base import Request

################################################################################

class RequestMsg(Msg):
    """ABC for messages describing request content."""

    @abstractmethod
    def decode(self, dec: DecoderP) -> 'Request': ...

    @property
    def is_async(self) -> bool:
        return False



# fmt: off

from asyncio import Future

from .__ import *
from .request_resource import ResourceRequest


__all__ = [
    'FutureRequest',
    'CancelFuture',
    'CompleteFuture',
]


################################################################################

if TYPE_CHECKING:
    from ..msg.all import FutureRequestMsg, CancelFutureMsg, CompleteFutureMsg

################################################################################

class FutureRequest(ResourceRequest):
    __slots__ = __match_args__ = 'future'

    @abstractmethod
    def encode(self, codec: 'EncoderP') -> 'FutureRequestMsg': ...


################################################################################

class CancelFuture(FutureRequest):
    __slots__ = __match_args__ = 'future',

    future: Future

    def __init__(self, future: Future):
        self.future = future

    def __execute__(self):
        self.future.cancel()

    def encode(self, enc) -> 'CancelFutureMsg':
        from ..msg.all import CancelFutureMsg
        return CancelFutureMsg(
            enc.enc_future(self.future)
        )

    def hook_key(self):
        return 'cancel_future'


################################################################################

class CompleteFuture(FutureRequest):
    __slots__ = __match_args__ = 'future', 'result'

    future: Future
    result: Result

    def __init__(self, future: Future, result: Result):
        self.future = future
        self.result = result

    def __execute__(self):
        result = self.result
        if result.is_ok:
            self.future.set_result(result.value)
        elif result.is_err:
            self.future.set_exception(result.error)
        else:
            pass

    def encode(self, enc) -> 'CompleteFutureMsg':
        from ..msg.all import CompleteFutureMsg, ResultMsg
        return CompleteFutureMsg(
            enc.enc_future(self.future),
            ResultMsg.encode(enc, self.result)
        )

    def hook_key(self):
        return 'complete_future'
